//******************************************************************************
//*                                  uBee512                                   *
//*       An emulator for the Microbee Z80 ROM, FDD and HDD based models       *
//*                                                                            *
//*                              6545 CRTC module                              *
//*                                                                            *
//*                   Copyright (C) 2007-2012 Stewart Kay                      *
//******************************************************************************
//
// Emulate the 6545 CRT for monochrome, standard colour and Premium/256TC
// colour models.
//
//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 ROM, FDD and HDD based models.
 *  Copyright (C) 2007-2012 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v5.0.0 - 3 August 2010, K Duckmanton
// - Refactored crtc_redraw(), crtc_draw_char(), crtc_update_cursor() and
//   crtc_setmonitor().  Foreground/background pixel values are now calculated
//   in crtc_setmonitor() (which is called when the video surface is created)
//   and the precalculated values are used in crtc_redraw() and crtc_draw_char().
//   crtc_draw_char() now uses the 'fast' pixel drawing function in video.c.
//
// v4.6.0 - 4 May 2010, S.J.Kay
// - Changes to crtc_regdump() to add binary output.
// - Renamed log_data() calls to log_data_1() and log_port() to log_port_1().
//
// v4.1.0 - 22 June 2009, S.J.Kay
// - Made improvements to the logging process.
//
// v4.0.0 - 6 June 2009, S.J.Kay
// - Changes made to incorporate a Z80 API.  Specific Z80 code used here
//   has been replaced with code usable by any Z80 emulator.
// - Masked off appended register values (upper 8 bits) from port values where
//   these were required.
// - The Vertical blanking period has been increased to 15% to produce
//   better sound with z80ex and when MWB v6.30e is used. This needs to be
//   calulated from the current CRT values rather than a constant.
//
// v3.1.0 - 6 December 2008, S.J.Kay
// - Constant used for the vertical blanking period has changed from 12/100
//   to 10/100 of the VSYNC period.
// - Added crtc_calc_vsync_freq() function to calculate the vertical sync
//   frequency from current CRTC 6545 values.
// - Changed dual intensity feature (--hint, now use --dint option) as was
//   incorrectly implemented.
// - Changed the way hardware flashing works, now also emulates early Premium
//   (main board v3) flash method.  See modelx.hwflash.
// - Added crtc_set_flash_rate() function to set the flash rate.
// - Changed the cursor blinking modes to use the correct blink rate.
// - Changed all printf() calls to use a local xprintf() function.
// - Modified crtc_configure() function to ignore SDL aspect if OpenGL mode.
// - Added OpenGL conditional code compilation using USE_OPENGL.
//
// v3.0.0 - 6 October 2008, S.J.Kay
// - Simplified the timer by introducing a time_get_ms() function.
// - Changed 'gui_' function prefix names to 'video_', added include video.h
//   and removed include gui.h,  other minor changes to function calls.
// - Random crashes (exits) on Windows when an OpenGL window was manually
//   resized was fixed by updating the display immediately in the
//   crtc_data_w() function when any of the CRTC registers 1, 6 or 9 or
//   changed, the code that had handled this in the crtc_update() function
//   has been removed. It should be remembered that this affects how big the
//   rendering buffer will be and so should be acted on immediately or make
//   sure that the values used for the screen set up are the ones used for
//   checking data boundaries.
// - Changed crtc_videochange() function to have better boundary checking.
//
// v2.8.0 - 18 August 2008, S.J.Kay
// - SDL video rendering has seen some major changes, SW and HW surfaces are
//   now supported along with 8, 16 or 32 bits per pixel modes.
// - Reversed engineered the contents of the standard IC 7 82s23 PROM resulting
//   in a new PROM table and changes to the order of values in the existing
//   standard colour tables.
// - Added a standard colour bit swap table for back ground colours to convert
//   from RGBrgb to BGRbgr order.
// - crtc.early_col_type is now crtc.std_col_type
// - Changed the default colour monitor type to RGBrgb for standard colour.
// - Added a crtc.flashbits counter that will determine when any flashing
//   video attribute bits are set, this should reduce host CPU time when
//   emulating alpha+ models.
//
// v2.7.0 - 3 July 2008, S.J.Kay
// - Premium colour table is now col_table_p.  Two colour tables have been
//   used for the early colour circuit as 2 monitor types were possible.
// - Code has been modified for the early colour model emulation.
// - Colour PROM table has been modified based on some photo results from
//   bgcol1.com and fgcol1.com Microbee colour test patterns.
// - Vertical blanking modified to return the correct duty cycle timing.
// - Added crtc_update() function, moved code over from the ubee512.c module.
// - Added structure emu_t and moved variables into it.
// - Added crtc_t structure. video member is able to be used to disable video
//   to the SDL window.
// - Added functions crtc_set_mon_table() and to set colour tables.
// - gui_videochange() is now gui_create_surface();
//
// v2.5.0 - 29 March 2008, S.J.Kay
// - Make 2MHZ model only able to access the first 2K of character ROM.
// - Added a configurable mon_table and monochrome selection to the
//   crtc_setmonitor() function.
// - Removed the const declaration from ic_82s23 array as the table values
//   are now able to be configured.
// - Removed the const declaration from col_table_2 array as the colours are
//   now able to be configured.
//
// v2.4.0 - 17 February 2008, S.J.Kay
// - Changed variables decleared in crtc_redraw() and crtc_redraw_char() to
//   static type that may result in faster execution.
//
// v2.3.0 - 8 January 2008, S.J.Kay
// - Added half intensity monochrome emulation for alpha+ models.
// - Added crtc_regdump() function to dump CRTC6545 registers.
// - Changes to crtc_setmonitor(), removed modelx.standard.
// - Added conditional test of (modelx.lpen) before the keystd_handler()
//   function is called in CRTC_DOSETADDR.
// - Added modio_t structure.
//
// v2.2.0 - 16 November 2007, S.J.Kay
// - Added original colour board emulation predating Premium and 256TC models.
// - Swapped RGB to BGR ordering in tables as this ordering was easier to work
//   out the colours for the original colour circuit.
//
// v2.1.0 - 27 October 2007, S.J.Kay
// - Reversed the full and half intensity colours in the col_table as these
//   were in the wrong order and re-adjusted the intensity levels to suit.
// - Added another intensity level to generate dark grey (was black).
// - Fiddled colour values (in crtc.h) to get better matching colours.
// - 6545 light pen keys emulation option added for 256tc model.
// - vblank status now in crtc_vblank() so that it can be accessed by the pio.c
//   module when used for port B bit 7.
// - Implement the modelx information structure.
//
// v2.0.0 - 14 October 2007, S.J.Kay
// - keyb_checkall() changed to keystd_checkall() and keyb_handler() changed
//   to keystd_handler().  These functions only called if not emulataing the
//   256TC model.
//
// v1.4.0 - 29 September 2007, S.J.Kay
// - Changes to Y scaling to achieve better aspect ratio.
// - Coding improvements to opt_monitor and opt_colour.  Some code tidy up.
// - Removed the non aplpha+ source build option.
//
// v1.0.0 - 14 August 2007, S.J.Kay
// - Added crtc_clock function to set xtal frequency and methods to be used for
//   the VBLANK status.  Now uses Z80 cycles or host timer to gererate VBLANK.
//   The host timer method is needed to keep VBLANK at 50 Hz in turbo mode.
// - Modifed the CRTC vertical blanking status method used.  The new method
//   now gives the correct delays needed for keyboard encoding and some games.
// - Implemented cursor modes (R10 bits 5,6), includes no cursor, steady cursor
//   and blinking cursor.
// - Writes to the CRTC registers (R1, R6, R9) will now cause the display to
//   be resized automatically as required.
// - Implemented the alpha+ inverse and flashing attribute bits.
// - Added standard and standard+colour model options.
// - Implemeted alpha+ (Premium) graphics in functions crtc_redraw and
//   crtc_redraw_char.
// - Added amber, green, black and white monitor options, default is colour.
//
// v0.0.0 - 21 June 2007, S.J.Kay
// - Start with "nanowasp" source distribution version 0.22. An emulator for
//   the microbee 128k. Copyright (C) 2000-2003  David G. Churchill.
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <SDL.h>

#include "ubee512.h"
#include "z80api.h"
#include "macros.h"
#include "support.h"
#include "crtc.h"
#include "keystd.h"
#include "vdu.h"
#include "video.h"

//==============================================================================
// structures and variables
//==============================================================================
static void crtc_calc_vsync_freq (void);
int crtc_update_cursor (void);

crtc_t crtc =
{
 .video = 1,
 .hdisp = 80,
 .vdisp = 25,
 .yscale = 1,
 .scans_per_row = 11,
 .vblank_method = 0,
 .monitor = 0,
 .std_col_type = 1,
 .flashrate = 4,
 .aspect = 2
};

static double vsync_freq;

static int cur_blink_rate_t1r32;
static int cur_blink_rate_t1r16;
static int cur_blink_rate_c1r32;
static int cur_blink_rate_c1r16;

static int cur_blink_last;
static int cur_blink;
static int cur_mode;
static int cur_pos;

static int flashvideo_last;

static int crtc_regs_data[32];
static int yscale_conf = 1;
static int vblank_divval;
static int vblank_cmpval;

static int htot;
static int vtot;
static int vtot_adj;
static int cur_start;
static int cur_end;
static int lpen_valid;
static int lpen;
static int reg;

static int monochrome;
static int mem_addr;
static int redraw;

static SDL_Rect rect_vdu =
{
 0,
 0,
 VDUWIDTH,
 VDUHEIGHT
};

static SDL_Rect rect_screen =
{
 VDUPOSX,
 VDUPOSY,
 VDUWIDTH,
 VDUHEIGHT
};

static SDL_PixelFormat *spf;
static uint32_t col_table[64];

#ifdef MINGW
#else
struct timeval tod_x;
#endif

static char *crtc_regs_names[] =
{
 "Horiz Total-1",
 "Horiz Displayed",
 "Horiz Sync Position",
 "VSYSNC, HSYNC Widths",
 "Vert Total-1",
 "Vert Total Adjust",
 "Vert Displayed",
 "Vert Sync Position",
 "Mode Control",
 "Scan Lines-1",
 "Cursor Start",
 "Cursor End",
 "Display Start Addr (H)",
 "Display Start Addr (L)",
 "Cursor Position (H)",
 "Cursor Position (L)",
 "Light Pen Reg (H)",
 "Light Pen Reg (L)",
 "Update Address Reg (H)",
 "Update Address Reg (L)"
 };

extern SDL_Surface *screen;
extern emu_t emu;
extern model_t modelx;
extern modio_t modio;
extern vdu_t vdu;
extern video_t video;

//==============================================================================
// User configurable monochrome monitor values
//==============================================================================

// indices into the colour table for each of the 4 "colours" in monochrome mode
// (background, high intensity background, foreground, high intensity foreground)
#define MONO_COLOUR_BG   0
#define MONO_COLOUR_BG_I 1
#define MONO_COLOUR_FG   2
#define MONO_COLOUR_FG_I 3

// amber on black colour table
const uint8_t monam_table[4][3] =
{
//       B              G              R
 {MONAM_BGB,    MONAM_BGG,    MONAM_BGR},
 {MONAM_BGB_I,  MONAM_BGG_I,  MONAM_BGR_I},
 {MONAM_FGB,    MONAM_FGG,    MONAM_FGR},
 {MONAM_FGB_I,  MONAM_FGG_I,  MONAM_FGR_I}
};

// green on black colour table
const uint8_t mongr_table[4][3] =
{
//       B              G              R
 {MONGR_BGB,    MONGR_BGG,    MONGR_BGR},
 {MONGR_BGB_I,  MONGR_BGG_I,  MONGR_BGR_I},
 {MONGR_FGB,    MONGR_FGG,    MONGR_FGR},
 {MONGR_FGB_I,  MONGR_FGG_I,  MONGR_FGR_I}
};

// black on white colour table
const uint8_t monbw_table[4][3] =
{
//       B              G              R
 {MONBW_BGB,    MONBW_BGG,    MONBW_BGR},
 {MONBW_BGB_I,  MONBW_BGG_I,  MONBW_BGR_I},
 {MONBW_FGB,    MONBW_FGG,    MONBW_FGR},
 {MONBW_FGB_I,  MONBW_FGG_I,  MONBW_FGR_I}
};

// white on black colour table
const uint8_t monwb_table[4][3] =
{
//       B              G              R
 {MONWB_BGB,    MONWB_BGG,    MONWB_BGR},
 {MONWB_BGB_I,  MONWB_BGG_I,  MONWB_BGR_I},
 {MONWB_FGB,    MONWB_FGG,    MONWB_FGR},
 {MONWB_FGB_I,  MONWB_FGG_I,  MONWB_FGR_I}
};

// user configurable colour table
uint8_t mon_table[4][3] =
{
//       B              G              R
 {MONUSR_BGB,    MONUSR_BGG,    MONUSR_BGR},
 {MONUSR_BGB_I,  MONUSR_BGG_I,  MONUSR_BGR_I},
 {MONUSR_FGB,    MONUSR_FGG,    MONUSR_FGR},
 {MONUSR_FGB_I,  MONUSR_FGG_I,  MONUSR_FGR_I}
};

//==============================================================================
// RGB analogue monitor colour values for the standard colour circuit.
//
// This table is for emulation of a colour monitor connected to X3, pins 11,
// 12 and 13.  There are 2 intensity levels per gun output colour.
//
// Reference: Microbee Technical manual 1986,  MB1217 schematic.
//
// The "xxbgrBGR" values determines the gun intensity (bgr) and colour (BGR)
// levels. This is the normal ordering in hardware so is implemented in the
// same way here.
//
// For a colour gun to be switched on requires the upper case gun bit
// position to be set (1).
//
// For the colour to be in high intensity the corresponding lower case gun
// letter bit position must also be set (1).
//
// For the colour to be in low intensity the corresponding lower case gun
// letter bit position must be clear (0).
//==============================================================================
static const uint8_t col_table_1[64][3] =
{
//   R     G     B              bgrBGR (00)
 {   0,    0,    0},    // 00 xx000000 black
 {   0,    0, C_HI},    // 04 xx000100 blue
 {   0, C_HI,    0},    // 02 xx000010 green
 {   0, C_HI, C_HI},    // 06 xx000110 cyan
 {C_HI,    0,    0},    // 01 xx000001 red
 {C_HI,    0, C_HI},    // 05 xx000101 magenta
 {C_HI, C_HI,    0},    // 03 xx000011 yellow
 {C_HI, C_HI, C_HI},    // 07 xx000111 grey

//   R     G     B              bgrBGR (01)
 {   0,    0,    0},    // 08 xx001000 black
 {   0,    0, C_FI},    // 12 xx001100 blue
 {   0, C_HI,    0},    // 10 xx001010 green
 {   0, C_HI, C_HI},    // 14 xx001110 cyan
 {C_FI,    0,    0},    // 09 xx001001 red
 {C_FI,    0, C_HI},    // 13 xx001101 magenta
 {C_FI, C_HI,    0},    // 11 xx001011 yellow
 {C_FI, C_HI, C_HI},    // 15 xx001111 grey

//   R     G     B              bgrBGR (02)
 {   0,    0,    0},    // 16 xx010000 black
 {   0,    0, C_HI},    // 20 xx010100 blue
 {   0, C_FI,    0},    // 18 xx010010 green
 {   0, C_FI, C_HI},    // 22 xx010110 cyan
 {C_HI,    0,    0},    // 17 xx010001 red
 {C_HI,    0, C_HI},    // 21 xx010101 magenta
 {C_HI, C_FI,    0},    // 19 xx010011 yellow
 {C_HI, C_FI, C_HI},    // 23 xx010111 grey

//   R     G     B              bgrBGR (03)
 {   0,    0,    0},    // 24 xx011000 black
 {   0,    0, C_HI},    // 28 xx011100 blue
 {   0, C_FI,    0},    // 26 xx011010 green
 {   0, C_FI, C_HI},    // 30 xx011110 cyan
 {C_FI,    0,    0},    // 25 xx011001 red
 {C_FI,    0, C_HI},    // 29 xx011101 magenta
 {C_FI, C_FI,    0},    // 27 xx011011 yellow
 {C_FI, C_FI, C_HI},    // 31 xx011111 grey

//   R     G     B              bgrBGR (04)
 {   0,    0,    0},    // 32 xx100000 black
 {   0,    0, C_FI},    // 36 xx100100 blue
 {   0, C_HI,    0},    // 34 xx100010 green
 {   0, C_HI, C_FI},    // 38 xx100110 cyan
 {C_HI,    0,    0},    // 33 xx100001 red
 {C_HI,    0, C_FI},    // 37 xx100101 magenta
 {C_HI, C_HI,    0},    // 35 xx100011 yellow
 {C_HI, C_HI, C_FI},    // 39 xx100111 grey

//   R     G     B              bgrBGR (05)
 {   0,    0,    0},    // 40 xx101000 black
 {   0,    0, C_FI},    // 44 xx101100 blue
 {   0, C_HI,    0},    // 42 xx101010 green
 {   0, C_HI, C_FI},    // 46 xx101110 cyan
 {C_FI,    0,    0},    // 41 xx101001 red
 {C_FI,    0, C_FI},    // 45 xx101101 magenta
 {C_FI, C_HI,    0},    // 43 xx101011 yellow
 {C_FI, C_HI, C_FI},    // 47 xx101111 grey

//   R     G     B              bgrBGR (06)
 {   0,    0,    0},    // 48 xx110000 black
 {   0,    0, C_FI},    // 52 xx110100 blue
 {   0, C_FI,    0},    // 50 xx110010 green
 {   0, C_FI, C_FI},    // 54 xx110110 cyan
 {C_HI,    0,    0},    // 49 xx110001 red
 {C_HI,    0, C_FI},    // 53 xx110101 magenta
 {C_HI, C_FI,    0},    // 51 xx110011 yellow
 {C_HI, C_FI, C_FI},    // 55 xx110111 grey

//   R     G     B              bgrBGR (07)
 {   0,    0,    0},    // 56 xx111000 black
 {   0,    0, C_FI},    // 60 xx111100 blue
 {   0, C_FI,    0},    // 58 xx111010 green
 {   0, C_FI, C_FI},    // 62 xx111110 cyan
 {C_FI,    0,    0},    // 57 xx111001 red
 {C_FI,    0, C_FI},    // 61 xx111101 magenta
 {C_FI, C_FI,    0},    // 59 xx111011 yellow
 {C_FI, C_FI, C_FI}     // 63 xx111111 white
};

//==============================================================================
// rgbRGB digital monitor colour values for the standard colour circuit.
//
// This table is for emulation of a rgbRGB colour monitor connected to X3,
// pins 3-8.
//
// Reference: Microbee Technical manual 1986,  MB1217 schematic(s).
//
// The "xxbgrBGR" values determines the gun intensity (bgr) and colour (BGR)
// levels. This is the normal ordering in hardware so is implemented in the
// same way here.
//==============================================================================
static const uint8_t col_table_2[64][3] =
{
//   R     G     B              bgrBGR (00)
 {   0,    0,    0},    // 00 xx000000 black
 {   0,    0, E_AA},    // 04 xx000100 blue
 {   0, E_AA,    0},    // 02 xx000010 green
 {   0, E_AA, E_AA},    // 06 xx000110 cyan
 {E_AA,    0,    0},    // 01 xx000001 red
 {E_AA,    0, E_AA},    // 05 xx000101 magenta
 {E_AA, E_AA,    0},    // 03 xx000011 yellow
 {E_AA, E_AA, E_AA},    // 07 xx000111 grey

//   R     G     B              bgrBGR (01)
 {   0,    0, E_55},    // 08 xx001000 black
 {   0,    0, E_FF},    // 12 xx001100 blue
 {   0, E_AA, E_55},    // 10 xx001010 green
 {   0, E_AA, E_FF},    // 14 xx001110 cyan
 {E_AA,    0, E_55},    // 09 xx001001 red
 {E_AA,    0, E_FF},    // 13 xx001101 magenta
 {E_AA, E_AA, E_55},    // 11 xx001011 yellow
 {E_AA, E_AA, E_FF},    // 15 xx001111 grey

//   R     G     B              bgrBGR (02)
 {   0, E_55,    0},    // 16 xx010000 black
 {   0, E_55, E_AA},    // 20 xx010100 blue
 {   0, E_FF,    0},    // 18 xx010010 green
 {   0, E_FF, E_AA},    // 22 xx010110 cyan
 {E_AA, E_55,    0},    // 17 xx010001 red
 {E_AA, E_55, E_AA},    // 21 xx010101 magenta
 {E_AA, E_FF,    0},    // 19 xx010011 yellow
 {E_AA, E_FF, E_AA},    // 23 xx010111 grey

//   R     G     B              bgrBGR (03)
 {   0, E_55, E_55},    // 24 xx011000 black
 {   0, E_55, E_FF},    // 28 xx011100 blue
 {   0, E_FF, E_55},    // 26 xx011010 green
 {   0, E_FF, E_FF},    // 30 xx011110 cyan
 {E_AA, E_55, E_55},    // 25 xx011001 red
 {E_AA, E_55, E_FF},    // 29 xx011101 magenta
 {E_AA, E_FF, E_55},    // 27 xx011011 yellow
 {E_AA, E_FF, E_FF},    // 31 xx011111 grey

//   R     G     B              bgrBGR (04)
 {E_55,    0,    0},    // 32 xx100000 black
 {E_55,    0, E_AA},    // 36 xx100100 blue
 {E_55, E_AA,    0},    // 34 xx100010 green
 {E_55, E_AA, E_AA},    // 38 xx100110 cyan
 {E_FF,    0,    0},    // 33 xx100001 red
 {E_FF,    0, E_AA},    // 37 xx100101 magenta
 {E_FF, E_AA,    0},    // 35 xx100011 yellow
 {E_FF, E_AA, E_AA},    // 39 xx100111 grey

//   R     G     B              bgrBGR (05)
 {E_55,    0, E_55},    // 40 xx101000 black
 {E_55,    0, E_FF},    // 44 xx101100 blue
 {E_55, E_AA, E_55},    // 42 xx101010 green
 {E_55, E_AA, E_FF},    // 46 xx101110 cyan
 {E_FF,    0, E_55},    // 41 xx101001 red
 {E_FF,    0, E_FF},    // 45 xx101101 magenta
 {E_FF, E_AA, E_55},    // 43 xx101011 yellow
 {E_FF, E_AA, E_FF},    // 47 xx101111 grey

//   R     G     B              bgrBGR (06)
 {E_55, E_55,    0},    // 48 xx110000 black
 {E_55, E_55, E_AA},    // 52 xx110100 blue
 {E_55, E_FF,    0},    // 50 xx110010 green
 {E_55, E_FF, E_AA},    // 54 xx110110 cyan
 {E_FF, E_55,    0},    // 49 xx110001 red
 {E_FF, E_55, E_AA},    // 53 xx110101 magenta
 {E_FF, E_FF,    0},    // 51 xx110011 yellow
 {E_FF, E_FF, E_AA},    // 55 xx110111 grey

//   R     G     B              bgrBGR (07)
 {E_55, E_55, E_55},    // 56 xx111000 black
 {E_55, E_55, E_FF},    // 60 xx111100 blue
 {E_55, E_FF, E_55},    // 58 xx111010 green
 {E_55, E_FF, E_FF},    // 62 xx111110 cyan
 {E_FF, E_55, E_55},    // 57 xx111001 red
 {E_FF, E_55, E_FF},    // 61 xx111101 magenta
 {E_FF, E_FF, E_55},    // 59 xx111011 yellow
 {E_FF, E_FF, E_FF}     // 63 xx111111 grey
};

//==============================================================================
// CGA Colour values for Premium, Teleterm and 256TC colour circuits
//==============================================================================
const uint8_t col_table_p[16][3] =
{
//   R     G     B                IBGR
 {   0,    0,    0},    // 00 xxxx0000 black
 {C_HI,    0,    0},    // 01 xxxx0001 red
 {   0, C_HI,    0},    // 02 xxxx0010 green
 {C_HI, C_LI,    0},    // 03 xxxx0011 brown
 {   0,    0, C_HI},    // 04 xxxx0100 blue
 {C_HI,    0, C_HI},    // 05 xxxx0101 magenta
 {   0, C_HI, C_HI},    // 06 xxxx0110 cyan
 {C_HI, C_HI, C_HI},    // 07 xxxx0111 light grey

 {C_LI, C_LI, C_LI},    // 08 xxxx1000 dark grey
 {C_FI, C_LI, C_LI},    // 09 xxxx1001 light red
 {C_LI, C_FI, C_LI},    // 10 xxxx1010 light green
 {C_FI, C_FI, C_LI},    // 11 xxxx1011 light yellow
 {C_LI, C_LI, C_FI},    // 12 xxxx1100 light blue
 {C_FI, C_LI, C_FI},    // 13 xxxx1101 light magenta
 {C_LI, C_FI, C_FI},    // 14 xxxx1110 light cyan
 {C_FI, C_FI, C_FI}     // 15 xxxx1111 white
};

//==============================================================================
// 82S123 PROM Colour table used on colour board (IC 7).
//
// The values found here have been reversed engineered by running some capture
// software on a standard colour model reading the values back in on the
// Microbee's parallel port.  The software took multiple samples to eliminate
// sync signals from the final result.
//
// See Microbee Disk System manaul section E-9.
//
// There are 26 unique colour combinations, or 27 if black is counted.
//
// Colours names with a leading '#' are colours that are not easy to describe
// and the colour description is intended as a guide only.
//==============================================================================
uint8_t ic_82s23[32] =
{
//                         nn Colour
 0x00,                  // 00 black
 0x09,                  // 01 blue
 0x12,                  // 02 green
 0x1b,                  // 03 cyan
 0x24,                  // 04 red
 0x2d,                  // 05 magenta
 0x36,                  // 06 yellow
 0x3f,                  // 07 white

 0x0b,                  // 08 #dark cyan
 0x13,                  // 09 #aqua
 0x0d,                  // 10 #purple
 0x2d,                  // 11 #pink
 0x16,                  // 12 #green
 0x26,                  // 13 #orange
 0x07,                  // 14 #grey
 0x3f,                  // 15 #white

 0x00,                  // 16 black
 0x01,                  // 17 blue II
 0x02,                  // 18 green II
 0x03,                  // 19 cyan II
 0x04,                  // 20 red II
 0x05,                  // 21 magenta II
 0x06,                  // 22 yellow II
 0x07,                  // 23 white II

 0x0f,                  // 24 #blue
 0x17,                  // 25 #green
 0x1f,                  // 26 #cyan
 0x27,                  // 27 #pink
 0x2f,                  // 28 #pink
 0x37,                  // 29 #beige
 0x07,                  // 30 #grey
 0x3f                   // 31 #white
};

//==============================================================================
// Standard colour Red/Blue bit swap for background colours.  This is required
// as the base colour values are arranged for the PROM lookup.
//
// See Microbee Disk System manaul section E-9.
//==============================================================================
uint8_t bg_standard_colour[8] =
{
//                         nn Colour
 B8(00000000),          // 00 black
 B8(00000100),          // 01 blue
 B8(00000010),          // 02 green
 B8(00000110),          // 03 cyan
 B8(00000001),          // 04 red
 B8(00000101),          // 05 magenta
 B8(00000011),          // 06 yellow
 B8(00000111)           // 07 grey/white
};

//==============================================================================
// CRTC set the rect_vdu and rect_screen (SDL_Rect) values using the current
// CRTC values.
//
//   pass: void
// return: void
//==============================================================================
static void crtc_SetRectValues (void)
{
 rect_vdu.x = VDUPOSX;  // x position of the upper-left corner of the rectangle
 rect_vdu.y = VDUPOSY;  // y position of the upper-left corner of the rectangle
 rect_vdu.w = crtc.hdisp * 8;           // The width of the rectangle

 rect_vdu.h = crtc.vdisp * crtc.scans_per_row * crtc.yscale;    // The height of the rectangle
 rect_screen.x = VDUPOSX; // x position of the upper-left corner of the rectangle
 rect_screen.y = VDUPOSY; // y position of the upper-left corner of the rectangle
 rect_screen.w = crtc.hdisp * 8;                // The width of the rectangle

 rect_screen.h = crtc.vdisp * crtc.scans_per_row * crtc.yscale; // The height of the rectangle
}

//==============================================================================
// CRTC set colour table
// Updates the cached SDL pixel values for each possible colour for the current
// screen surface.
//
//   pass: void
// return: void
//==============================================================================
void crtc_setcolourtable()
{
 int i;
 const uint8_t (*coltable)[3];

 if (monochrome)
    {
     /* For monochrome models we use the first 4 entries in col_table.
        The entries are numbered like so:

        0: low intensity background
        1: high intensity background
        2: low intensity foreground
        3: high intensity foreground
        
        Note that if the 1/2 intensity monochrome hardware isn't being
        modeled, the pixel value for high and low intensity will be the same
     */

     // set the colour scheme according to the monitor type
     switch (crtc.monitor)
        {
         default:
         case 0 :
            // colour (default)
            coltable = mongr_table;
            break;
         case 1 :
            // amber on black
            coltable = monam_table;
            break;
         case 2 :
            // green on black
            coltable = mongr_table;
            break;
         case 3 :
            // black on white
            coltable = monbw_table;
            break;
         case 4 :
            // white on black
            coltable = monwb_table;
            break;
         case 5 :
            // user configurable monochrome
            coltable = (const uint8_t(*)[3])mon_table;
            break;
        }

     for (i = 0; i < 4; ++i)
        {
         col_table[i] = SDL_MapRGB(spf, coltable[i][2], coltable[i][1], coltable[i][0]);
        }

     if (!modelx.alphap || !modelx.halfint)
        {
         col_table[MONO_COLOUR_BG_I] = col_table[MONO_COLOUR_BG];
         col_table[MONO_COLOUR_FG] = col_table[MONO_COLOUR_FG_I];
        }
    }
 else
    {
     // all colour models
     if (modelx.alphap)
        {
         // premium/256TC
         coltable = col_table_p;
         i = sizeof(col_table_p) / sizeof(col_table_p[0]);
        }
     else if (crtc.std_col_type == 0)
        {
         coltable = col_table_1;
         i = sizeof(col_table_1) / sizeof(col_table_1[0]);
        }
     else
        {
         coltable = col_table_2;
         i = sizeof(col_table_2) / sizeof(col_table_2[0]);
        }
     while (i--)
        {
         col_table[i] = SDL_MapRGB(spf,
                                   coltable[i][0],
                                   coltable[i][1],
                                   coltable[i][2]);
        }
    }
}

//==============================================================================
// CRTC set monitor
//
//   pass: void
// return: void
//==============================================================================
void crtc_setmonitor (void)
{
 monochrome = (modelx.colour == 0) || (crtc.monitor);
 crtc_setcolourtable();
 SDL_FillRect(screen, &rect_vdu, col_table[0]); // default background colour
}

//==============================================================================
// CRTC Change Video
//
//
//   pass: void
// return: int                  0 if success, -1 if error
//==============================================================================
static int crtc_videochange (void)
{
 int crt_w;
 int crt_h;

 crt_w = crtc.hdisp * 8;
 crt_h = crtc.vdisp * crtc.scans_per_row;

 if ((crt_w) && (crt_w <= 720) && (crt_h) && (crt_h <= 600))
    {
#ifdef USE_OPENGL
     if (video.type == VIDEO_GL)
        {
         if ((video.maximised) && (! video.fullscreen))
            video_create_surface(crt_w, crt_h, video.maximised_w, video.maximised_h);
         else
            video_create_surface(crt_w, crt_h, video.last_win_w, video.last_win_h);
        }
     else
#endif
        {
         if (crtc.hdisp < 50)           // for programs like Videotex (40 Col)
            crtc.yscale = 1;
         else
            crtc.yscale = yscale_conf;

         crtc_SetRectValues();

         video_create_surface(crt_w, crt_h, 0, 0);
        }

     spf = screen->format;
     crtc_setmonitor();
     crtc_set_redraw();
     crtc_redraw();
     video_renderer();

     crtc.resized = 0;          // clear the resized flag

     return 0;
    }

 return -1;
}

//==============================================================================
// CRTC Initialise
//
//   pass: void
// return: int                  0 if sucess, -1 if error
//==============================================================================
int crtc_init (void)
{
 spf = screen->format;

 crtc_setmonitor();
 return 0;
}

//==============================================================================
// CRTC de-initialise
//
//   pass: void
// return: int                  0
//==============================================================================
int crtc_deinit (void)
{
 return 0;
}

//==============================================================================
// CRTC reset
//
//   pass: void
// return: int                  0
//==============================================================================
int crtc_reset (void)
{
 reg = 0;

 return 0;
}


//==============================================================================
// CRTC vblank status
//
// The Vertical blanking status is generated from the Z80 clock cycles that
// have elapsed or the host timer depending on the mode required.
//
//   pass: void
// return: int                  vblank status (in bit 7)
//==============================================================================
int crtc_vblank (void)
{
 uint64_t cycles_now;

 if (crtc.vblank_method == 0)
    {
     cycles_now = z80api_get_tstates();
     if ((cycles_now % vblank_divval) < vblank_cmpval)
        return B8(10000000);
    }
 else
    {
     if ((time_get_ms() / 10) & 1)      // div 10mS (100Hz)
        return B8(10000000);            // return true at a 50Hz rate
    }

 return 0;
}

//==============================================================================
// CRTC status
//
//   pass: uint16_t port
//         struct z80_port_read *port_s
// return: uint16_t             status
//==============================================================================
uint16_t crtc_status (uint16_t port, struct z80_port_read *port_s)
{
 int status = 0x80;                     // always set update strobe

 if ((modelx.lpen) && (! lpen_valid))
    keystd_checkall();

 if ((modelx.lpen) && (lpen_valid))     // NB: this is not an else because
    status |= 0x40;                     // keystd_checkall might set lpen_valid

 if (crtc_vblank())
    status |= 0x20;

 if (modio.crtc)
    log_port_1("crtc_status", "status", port, status);

 return status;
}

//==============================================================================
// CRTC Light Pen
//
//   pass: int addr
// return: void
//==============================================================================
void crtc_lpen (int addr)
{
 if (! lpen_valid)
    {
     lpen_valid = 1;
     lpen = addr;
     if (modio.crtc)
     log_data_1("crtc_lpen", "addr", addr);
    }
}

//==============================================================================
// Set CRTC reg address - Port function
//
//   pass: uint16_t port
//         uint8_t data
//         struct z80_port_write *port_s
// return: void
//==============================================================================
void crtc_address (uint16_t port, uint8_t data, struct z80_port_write *port_s)
{
 reg = data & 0x1F;

 if (modio.crtc)
    log_port_1("crtc_address", "data", port, data);
}

//==============================================================================
// Read CRTC register data - Port function
//
//   pass: uint16_t port
//         struct z80_port_read *port_s
// return: uint16_t register    register data
//==============================================================================
uint16_t crtc_data_r (uint16_t port, struct z80_port_read *port_s)
{
 uint16_t val;

 switch (reg)
    {
     case CRTC_CUR_POS_H:
        val = (cur_pos >> 8) & 0x3F;
        break;
     case CRTC_CUR_POS_L:
        val = cur_pos & 0xFF;
        break;

     case CRTC_LPEN_H:
        lpen_valid = 0;
        val = (lpen >> 8) & 0x3F;
        break;
     case CRTC_LPEN_L:
        lpen_valid = 0;
        val = lpen & 0xFF;
        break;

     default:
        val = 0xFFFF;
    }

 if (modio.crtc)
    log_port_2("crtc_data_r", "reg", "val", port, reg, val);

 return val;
}

//==============================================================================
// Write CRTC register data - Port function
//
// crtc_redraw, and crtc_redraw_char functions
// -------------------------------------------
// uses: crtc.hdisp, crtc.vdisp, crtc.disp_start, crtc.scans_per_row,
//       cur_start, cur_end, cur_pos, cur_mode
//
// Setting up the display resolution
// ---------------------------------
// The CRTC_HDISP data is placed in crtc.hdisp.
// The CRTC_VDISP data is placed in crtc.vdisp.
// The CRTC_SCANLINES data is placed in crtc.scans_per_row
//
// The X resolution is determined from crtc.hdisp * 8 The Y resolution is
// determined from crtc.vdisp * crtc.scans_per_row
//
//   pass: uint16_t port
//         uint8_t data
//         struct z80_port_write *port_s
// return: void
//==============================================================================
void crtc_data_w (uint16_t port, uint8_t data, struct z80_port_write *port_s)
{
 int old_curpos;

 if (modio.crtc)
    log_port_1("crtc_data_w", "data", port, data);

 crtc_regs_data[reg] = data;

 switch (reg)
    {
     case CRTC_HTOT:            // R0
        htot = (data & 0xFF) + 1;
        crtc_calc_vsync_freq();
        break;
     case CRTC_HDISP:           // R1
        if (crtc.hdisp != (data & 0xFF))        // if the value has really changed
           {
            crtc.hdisp = data & 0xFF;
            crtc.resized = 1;
           }
        crtc_calc_vsync_freq();
        break;

     case CRTC_VTOT:            // R4
        vtot = (data & 0x7F) + 1;
        crtc_calc_vsync_freq();
        break;
     case CRTC_VTOT_ADJ:        // R5
        vtot_adj = data & 0x1F;
        crtc_calc_vsync_freq();
        break;

     case CRTC_VDISP:           // R6
        if (crtc.vdisp != (data & 0x7F))        // if the value has really changed
           {
            crtc.vdisp = data & 0x7F;
            crtc.resized = 1;
           }
        break;

     case CRTC_SCANLINES:       // R9
        if (crtc.scans_per_row != ((data & 0x1F) + 1)) // if the value has really changed
           {
            crtc.scans_per_row = (data & 0x1F) + 1;
            crtc.resized = 1;
           }
        crtc_calc_vsync_freq();
        break;

     case CRTC_CUR_START:       // R10
        cur_start = data & 0x1F;
        cur_mode = (data >> 5) & 0x03;
        crtc_update_cursor();
        crtc_redraw_char(cur_pos, 0);
        break;
     case CRTC_CUR_END:         // R11
        cur_end = data & 0x1F;
        crtc_redraw_char(cur_pos, 0);
        break;

     case CRTC_DISP_START_H:    // R12
        crtc.disp_start &= 0xFF;
        crtc.disp_start |= (data & 0x3F) << 8;
        break;
     case CRTC_DISP_START_L:    // R13
        crtc.disp_start &= 0x3F00;
        crtc.disp_start |= data & 0xFF;
        break;

     case CRTC_CUR_POS_H:       // R14
        old_curpos = cur_pos;
        cur_pos &= 0xFF;
        cur_pos |= (data & 0x3F) << 8;
        crtc_redraw_char(old_curpos, 0);
        crtc_redraw_char(cur_pos, 0);
        break;
     case CRTC_CUR_POS_L:       // R15
        old_curpos = cur_pos;
        cur_pos &= 0x3F00;
        cur_pos |= data & 0xFF;
        crtc_redraw_char(old_curpos, 0);
        crtc_redraw_char(cur_pos, 0);
        break;

     case CRTC_SETADDR_H:       // R18
        mem_addr = (mem_addr & 0x00FF) | ((data & 0x3F) << 8);
        break;
     case CRTC_SETADDR_L:       // R19
        mem_addr = (mem_addr & 0x3F00) | (data & 0xFF);
        break;

     case CRTC_DOSETADDR:       // R31
        if (modelx.lpen)
           keystd_handler(mem_addr);
        break;
   }

  if (crtc.resized)
     crtc_videochange();        // resets crtc.resized value
}

//==============================================================================
//
//   pass: int maddr            address to be redrawn
//         char *inverse        pointer to the inverse flag
// return: char *               pointer to character data for that character
//                              cell
//==============================================================================
static inline
char *crtc_char_data(int maddr, uint8_t *inverse)
{
 char ch, *data;
 uint8_t pcgbank, attrib;
 int romofs;

 if (emu.model != MOD_2MHZ)
    romofs = (maddr & 0x2000) >> 2;
 else
    romofs = 0;

 maddr &= 0x7ff;

 ch = vdu.scr_ram[maddr];
 *inverse = 0;

 if (crtc.extendram)
    {
     attrib = vdu.att_ram[maddr];
     pcgbank = attrib & 0x0F;

     if (attrib & B8(11000000))    // if inverse or flash attribute bit set
        {
         if (modelx.hwflash == 1)
            {
             // Premium (v4 mb), 256tc/tterm models flash by alternating between normal and blank
             if ((attrib & B8(10000000)) && (crtc.flashvideo)) // if flash attribute bit set
                ch = 32; // use blank instead when flashing
             if (attrib & B8(01000000)) // if inverse attribute bit set
                *inverse = 0xFF;
            }
         else
            {
             // Premium models (v3 mb) flash by alternating between normal and inverted
             if ((attrib & B8(10000000)) && (crtc.flashvideo)) // if flash attribute bit set
                *inverse = 0xFF; // use inverse instead when flashing
            }
        }

     if (ch & 0x80)                     // selects PCG RAM
        data = vdu.pcg_ram + pcgbank * 0x0800 + 16 * (ch & 0x7F);
     else
        data = vdu.chr_rom + 16 * (ch & 0x7F) + romofs;
    }
 else
    {
     if (ch & 0x80)                     // selects PCG RAM
        data = vdu.pcg_ram + 16 * (ch & 0x7F);
     else
        data = vdu.chr_rom + 16 * (ch & 0x7F) + romofs;
    }

 return data;
}

//==============================================================================
//
//   pass: int maddr            address to be redrawn (must be normalised)
//         int *fgc             address of foreground colour variable
//         int *bgc             address of background colour variable
// return: void
//                              *fgc and *bgc are updated
//==============================================================================
static inline
void crtc_char_colour(int maddr, uint32_t *xfgc, uint32_t *xbgc)
{
 uint8_t colour;
 uint8_t fgc, bgc;

 if (! monochrome)
    {
     colour = vdu.col_ram[maddr];
     if (modelx.colour == MODCOL1)
        {
         fgc = ic_82s23[colour & B8(00011111)];
         bgc = (bg_standard_colour[(crtc.colour_cont & B8(00001110)) >> 1] << 3) |
               (bg_standard_colour[(colour & B8(11100000)) >> 5]);
        }
     else
        {
         fgc = (colour & 0x0F);
         bgc = (colour >> 4);
        }
    }
 else
    {
     fgc = 2;
     bgc = 0;
     if ((modelx.alphap) && (modelx.halfint))
        {
         if (vdu.col_ram[maddr] & B8(00001000))
            fgc++;
         if (vdu.col_ram[maddr] & B8(10000000))
            bgc++;
        }
    }
 *xfgc = col_table[fgc];
 *xbgc = col_table[bgc];
}

//==============================================================================
// Redraw one character position.
//
//   pass: char *data           character data
//         int x, y             position of upper left corner, pixels
//         uint8_t inverse      inverse byte
//         uint8_t cursor       cursor byte
//         uint32_t xfgc        foreground colour
//         uint32_t xbgc        background colour
//==============================================================================
static inline
void crtc_draw_char_low(char *data, int x, int y,
                        uint8_t inverse, uint8_t cursor,
                        uint32_t xfgc, uint32_t xbgc)
{
 int i, j;
 uint32_t bit_colour;
 uint8_t ch;

 for (i=0; i < crtc.scans_per_row; i++)
    {
     ch = *(data++) ^ inverse;

     if (cursor && (i >= cur_start) && (i <= cur_end))
        ch ^= cursor;

     for (j=0; j < 8; j++)
        {
         bit_colour = (ch & 0x80) ? xfgc : xbgc;
         video_putpixel_fast(x+j, y, bit_colour);
         if (crtc.yscale == 2)
            video_putpixel_fast(x+j, y+1, bit_colour);
         ch <<= 1;
        }
     y += crtc.yscale;
    }
}

//==============================================================================
// redraw one screen address character position.
//
//   pass: int addr             address to be redrawn
//         int dostdio          1 if stdout should be used
// return: void
//==============================================================================
void crtc_redraw_char (int addr, int dostdout)
{
 int x, y;
 int offset;
 int maddr;
 uint8_t cursor, inverse;
 char *data;
 uint32_t xfgc;
 uint32_t xbgc;

 if ((crtc.hdisp == 0) || (! crtc.video))
    return;

 offset = (addr - crtc.disp_start) & 0x7ff;
 x = offset % crtc.hdisp;                       // character column
 y = offset / crtc.hdisp;                       // character row

 if (y >= crtc.vdisp)                           // character position is offscreen
    return;

 maddr = (crtc.disp_start + offset) & 0x3fff;
 data = crtc_char_data(maddr, &inverse);
 crtc_char_colour(maddr & 0x07ff, &xfgc, &xbgc);
 cursor = (maddr == cur_pos) ? cur_blink : 0x00;
 SDL_LockSurface(screen);
 crtc_draw_char_low(data,
                    x * 8, y * crtc.yscale * crtc.scans_per_row,
                    inverse, cursor, xfgc, xbgc);
 SDL_UnlockSurface(screen);
 crtc.update = 1;
}

//==============================================================================
// Set the redraw flag so that the next crtc_redraw function call is carried
// out.
//
//   pass: void
// return: void
//==============================================================================
void crtc_set_redraw (void)
{
 redraw = 1;
}

//==============================================================================
// Update the whole screen area if redraw is 1, or only PCG characters that
// were changed if redraw is 2.
//
//   pass: void
// return: void
//==============================================================================
void crtc_redraw (void)
{
 int i, j, x, y, l;
 int maddr;
 uint8_t cursor, inverse;
 uint32_t xfgc, xbgc;

 if ((! redraw) || (! crtc.video))
    return;
 SDL_LockSurface(screen);
 maddr = crtc.disp_start;
 l = crtc.yscale * crtc.scans_per_row;
 for (y = 0, i = 0; i < crtc.vdisp; i++, y += l)
    for (x = 0, j = 0; j < crtc.hdisp; j++, x += 8)
       {
        char *data;
        maddr &= 0x3fff;
        data = crtc_char_data(maddr, &inverse);
        crtc_char_colour(maddr & 0x7ff, &xfgc, &xbgc);
        cursor = (maddr == cur_pos) ? cur_blink : 0x00;
        crtc_draw_char_low(data, x, y,
                           inverse, cursor,
                           xfgc, xbgc);
        maddr++;
       }
 SDL_UnlockSurface(screen);
 crtc.update = 1;
 redraw = 0;
}

//==============================================================================
// CRTC update cursor.
//
// This function updates the state of the cursor.  It returns true if the
// cursor state has changed.
//
//   pass: void
// return: int
//==============================================================================
int crtc_update_cursor(void)
{
 // Determine the current status for the CRTC blinking cursor and refresh it
 // if this has changed.  The method used here depends on if turbo mode is
 // used. If turbo mode is used then Z80 execution speed will not be known as
 // no delays will be inserted, if not turbo then the rate must be determined
 // by the Z80 cycle count to achieve smooth results.
 switch (cur_mode)
    {
     default:
     case 0:
        cur_blink_last = cur_blink = 0xff; /* cursor always displayed */
        break;
     case 1:
        cur_blink_last = cur_blink = 0x00; /* cursor off */
        break;
     case 2:
        // blinking at 1/32 field rate
        cur_blink =
           (((emu.turbo) ?
             (time_get_ms() / cur_blink_rate_t1r32) :
             (emu.z80_cycles / cur_blink_rate_c1r32)) & 0x01) ? 0xff: 0x00;
        break;
     case 3:
        // blinking at 1/16 field rate
        cur_blink =
           (((emu.turbo) ?
             (time_get_ms() / cur_blink_rate_t1r16) :
             (emu.z80_cycles / cur_blink_rate_c1r16)) & 0x01) ? 0xff: 0x00;
        break;
    }
 if (cur_blink != cur_blink_last)
    {
     cur_blink_last = cur_blink;
     return 1;                  // changed
    }
 else
    return 0;
}


//==============================================================================
// CRTC update.
//
// Updates the CRTC periodically.
//
//   pass: void
// return: void
//==============================================================================
void crtc_update (void)
{
 if (crtc_update_cursor())
    crtc_redraw_char(cur_pos, 0);
// Determine the current state of the alpha+ flashing video and refresh it
// if this has changed.
 if (crtc.extendram)                    // only if extended RAM selected
    {
     if (emu.turbo)
        {
         if ((time_get_ms() / crtc.flashvalue_t) & 0x01)
            crtc.flashvideo = modelx.hwflash;
         else
            crtc.flashvideo = 0;
        }
     else
        {
         if ((emu.z80_cycles / crtc.flashvalue_c) & 0x01)
            crtc.flashvideo = modelx.hwflash;
         else
            crtc.flashvideo = 0;
        }
     if (crtc.flashvideo != flashvideo_last)
        {
         flashvideo_last = crtc.flashvideo;
         if (crtc.flashbits) // if we have some flashing attribute bits
            crtc_set_redraw();  // force crtc to redraw screen
        }
    }
}

//==============================================================================
// CRTC register dump
//
// Dump the contents of the crtc registers.
//
//   pass: void
// return: void
//==============================================================================
void crtc_regdump (void)
{
 int i;
 char s[17];

// CRTC_CUR_POS_H
 crtc_regs_data[14] = (cur_pos >> 8) & 0x3F;

// CRTC_CUR_POS_L
 crtc_regs_data[15] = cur_pos & 0xFF;

// CRTC_LPEN_H
 crtc_regs_data[16] = (lpen >> 8) & 0x3F;

// CRTC_LPEN_L
 crtc_regs_data[17] = lpen & 0xFF;

 xprintf("\n");
 xprintf("6545 CRTC Registers                Hex  Dec    Binary\n");
 xprintf("------------------------------------------------------\n");

 for (i = 0; i < 20; i++)
    xprintf("0x%02x (%02dd) %-22s  %02x %5d %10s\n", i, i, crtc_regs_names[i],
    crtc_regs_data[i], crtc_regs_data[i], i2b(crtc_regs_data[i],s));
}

//==============================================================================
// CRTC set monochrome colour table.
//
// Sets the RGB values for foreground, background full and half intensities.
//
//   pass: int pos
//         int col
// return: void
//==============================================================================
void crtc_set_mon_table (int pos, int col)
{
 mon_table[pos/3][pos%3] = col;
}

//==============================================================================
// CRTC configure.
//
// Determines the displayed aspect ratio to use.
//
//   pass: int aspect
// return: void
//==============================================================================
void crtc_configure (int aspect)
{
 if (! video.type != VIDEO_GL)
    {
     yscale_conf = aspect;
     crtc.yscale = aspect;
    }
}

//==============================================================================
// CRTC set hardware flash rate.
//
// Sets the flash rate for the alpha+ flashing attribute bit.  The flash
// rate is determined by IC60 a dual 4-bit binary counter, 4 link settings
// (W6x) and the VSYNC signal (typ 50Hz).  The settings for a V4 main board
// are as follows:
//
// Number   74LS393   Link      Rate (milliseconds)
// 0        1QA                 20
// 1        1QB                 40
// 2        1QC                 80
// 3/8      1QD       W61 A-B   160
// 4/9      2QA       W62 A-B   320
// 5/10     2QB       W63 A-B   640
// 6/11     2QC       W64 A-B   1280
// 7        2QD                 2560
//
// Four link settings (W61-W64) are provided on the main board, other values
// are possible by connecting to other pins.
//
// NOTE: Version 3 boards have 1280mS for W63 and 640mS for W64.
//
// This function should be called when the flash rate option is used and
// after the CPU clock speed is set.
//
//   pass: int n                flash rate number 0-11
// return: int                  0 if sucess, -1 if error
//==============================================================================
int crtc_set_flash_rate (int n)
{
 double t;

 if ((n < 0) || (n > 11))
    return -1;

 if (n < 8)
    crtc.flashrate = n;
 else
    crtc.flashrate = (n - 8) + 3;

 t = (1.0 / vsync_freq) * (1 << n);

 crtc.flashvalue_c = (int)(emu.cpuclock * t);
 crtc.flashvalue_t = (int)(t * 1000.0);

 return 0;
}

//==============================================================================
// CRTC clock calculations.
//
// VERTICAL BLANKING
// -----------------
// The Vertical blanking status is generated from the Z80 clock cycles that
// have elapsed or the host timer depending on the mode required.
//
// The Vertical blanking period is emulated to produce about a 15% (was 10%)
// on duty cycle. The emulated blanking frequency is ~50 frames per second
// for normal usage or can be a proportional value calculated from the CPU
// clock frequency.
//
// The vertical blanking period can not be derived from the host timer in
// the manner expected because the timer is continuous and the Z80 CPU
// emulation is achieved in frames, so a basic 50% duty is returned for this
// mode.  This is mainly intended for when running high speed emulation to
// keep key repeat speed usable.
//
// The vertical blanking is commonly used for keyboard encoding delays and for
// delays used in some games.
//
// CURSOR BLINKING
// ---------------
// blinking time (1/16 field rate) = 16 / vsync_freq
// blinking time (1/32 field rate) = 32 / vsync_freq
//
//   pass: int cpuclock         CPU clock frequency in Hz.
// return: void
//==============================================================================
void crtc_clock (int cpuclock)
{
 vblank_divval = (int)(cpuclock / vsync_freq);  // 67500 if 50Hz
 vblank_cmpval = (int)(vblank_divval * (15.0/100.0));

 // blinking at 1/32 field rate
 cur_blink_rate_t1r32 = (int)((32.0 / vsync_freq) * 1000);
 cur_blink_rate_c1r32 = (int)(cpuclock * (32.0 / vsync_freq));

 // blinking at 1/16 field rate
 cur_blink_rate_t1r16 = (int)((16.0 / vsync_freq) * 1000);
 cur_blink_rate_c1r16 = (int)(cpuclock * (16.0 / vsync_freq));

 crtc_set_flash_rate(crtc.flashrate);
}

//==============================================================================
// CRTC VSYNC calculations.
//
// Calculate the vertical sync frequency.  2MHz models use 12MHz and all
// others use a 13.5MHz crystal.
//
//   pass: void
// return: void
//==============================================================================
static void crtc_calc_vsync_freq (void)
{
 double vdu_xtal;

 if (emu.model == MOD_2MHZ)
    vdu_xtal = 12.0E+6;
 else
    vdu_xtal = 13.5E+6;

 if (htot && vtot && crtc.scans_per_row)
    {
     vsync_freq = (vdu_xtal / (htot * 8)) / (vtot * crtc.scans_per_row + vtot_adj);
     // adjust everything that relies on the VSYNC frequency
     crtc_clock(emu.cpuclock);
    }

 // avoid divide by 0 errors
 if (vsync_freq < 5.0)
    vsync_freq = 1.0;

#if 0
 xprintf("crtc_sync: vsync_freq=%f\n", vsync_freq);
#endif
}
