//******************************************************************************
//*                                  uBee512                                   *
//*       An emulator for the Microbee Z80 ROM, FDD and HDD based models       *
//*                                                                            *
//*                                 VDU module                                 *
//*                                                                            *
//*                   Copyright (C) 2007-2012 Stewart Kay                      *
//******************************************************************************
//
// Emulates the graphics hardware for Standard and Premium models (Alpha+)
//
// - Character ROM
// - Screen RAM
// - PCG RAM
// - Colour RAM
// - Attribute RAM
//
//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 ROM, FDD and HDD based models.
 *  Copyright (C) 2007-2012 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v4.6.0 - 4 May 2010, S.J.Kay
// - Changes to vdu_init() function to clear the colour RAM with 0s to
//   prevent the checker board pattern on start-up. This is only a temporary
//   work around and the real solution may need some type of 6545
//   initialisation delay to make it work like a real model.
// - Renamed log_port() calls to log_port_1() and log_data() to log_port_0().
//
// v4.1.0 - 22 June 2009, S.J.Kay
// - Code speed up for video read/write functions.
// - Made improvements to the logging process.
// - Loading of character ROM and colour PROM moved to new roms.c module.
//
// v4.0.0 - 7 June 2009, S.J.Kay
// - Changes made to incorporate a Z80 API.  Specific Z80 code used here
//   has been replaced with code usable by any Z80 emulator.
// - Completely overhauled memory management method, there is no longer any
//   banked memory swapping.
// - Added crtc_set_redraw() call to vdu_lvdat_w() function to update state
//   of display (i.e Telcom v3 problem fix introduced in v3.1.0) as the
//   vdu_switchvideo_in() function is now removed.
// - Set some vdu.scr_ram[], vdu.pcg_ram[] locations to 0xff as some
//   Microbee applications expect RAM to come up with certain values in some
//   locations.  (Above 2 fixes submitted by workerbee)
// - Masked off appended register values (upper 8 bits) from port values where
//   these were required.
//
// v3.1.0 - 22 April 2009, S.J.Kay
// - Removed all occurrences of console_output() function calls.
// - Fixed a display refresh problem introduced in v3.0.0/1 when running in
//   standard ROM model monochrome mode and Telcom versions 1.1 or 1.2.  The
//   top status line was incorrect, see vdu_colcont_w() function notes.
// - Fixed a display refresh problem introduced in v2.8.0. This problem was
//   noticable in Telcom v3 on PJB system disk and some other applications. A
//   crtc_set_redraw() call was added to vdu_switchvideo_in() function.
// - Changed all printf() calls to use a local xprintf() function.
//
// v3.0.0 - 21 September 2008, S.J.Kay
// - Recoded the vdu_colcont_w() function to test RGB intensity bits and VDU
//   PCG/colour RAM selection bits seperately, this prevents a change in
//   either leading to unnecessary time consuming updates.
//
// v2.8.0 - 15 August 2008, S.J.Kay
// - Standard BG colour intensity was not updating after a write to the colour
//   port in vdu_colcont_w() function.  Added crtc_set_redraw() call.
// - Added a crtc.flashbits counter that will determine when any flashing
//   video attribute bits are set, this will reduce host CPU time when
//   emulating alpha+ models.
// - vdu_vidmem_w() function now tests current video state before making calls
//   to crtc_set_redraw() and crtc_redraw_char() functions to reduce the host
//   CPU time.
// - Cast uint8_t ic_82s23[] as char* in memmap_loadrom function call to
//   eliminate a sign warning.
//
// v2.7.0 - 4 June 2008, S.J.Kay
// - New open_file() function in function.c module is now used by the
//   memmap_loadrom() function to determine the path to the ROMs.  Added full
//   file path return variable.
// - Added structures emu_t, crtc_t, memmap_t and moved variables into it.
//
// v2.5.0 - 29 March 2008, S.J.Kay
// - Added code for --colprom option to overide early built in colour
//   82s123 IC 7 PROM data.
// - Character ROM may now be a 2K (2716) or 4K (2732) image type.
// - The character ROM can now be specified as a user option.
// - Implement the modelc structure.
//
// v2.4.0 - 17 February 2008, S.J.Kay
// - VDU_MASK0 and VDU_MASK3 removed and used modelx.vdu instead.  The VDU
//   RAM (colour RAM) mask of 3 did not work as expected when software has
//   port 0x1c set to a non zero PCG bank and VDU selected with the
//   intention of accessing the standard VDU memory. It would appear
//   Microbee removed IC14 (Premium) and linked pins 2 and 23 of the 6264 to
//   0V effectively making it into a 2K RAM part so that any software that
//   did not do an out 0x1c,0x80 before attempting to access the standard 2K
//   VDU would still work correctly.
// - Last version incorrectly claims all alpha+ models shipped with 8K for
//   each VDU RAM.  Early models had 2K later increased to 8K.
//
// v2.3.0 - 23 January 2008, S.J.Kay
// - VDU and PCG mask variables and vdu_configure() function removed.  The
//   number of PCG banks is now handled correctly.  The VDU mask is now a
//   constant as all alpha+ models had 8K screen, attribute and colour RAM.
//   The VDU_MASK0 define is set to 0 for screen and attribute RAM, a value
//   of 3 is not working, 256TC boot ROM has incorrect graphics and locking
//   is evident when it's 3. VDU_MASK3 define is used for extended colour
//   RAM and appears to work correctly.
// - ROM Basic 5.22e colour failure mystery solved and fixed by emulating
//   static memory pattern.  Removed VDU init code from memory vdu_reset()
//   function.
// - Many changes made to the colour and alpha+ graphics read/write and the
//   video switching in/out functions.
// - Made changes to the vdu_switchvideo_in() function to suit all model types.
// - Changes to the vdu_lvdat_r() and vdu_lvdat_w() functions due to changes
//   made in z80.c
// - Added modio_t structure.
//
// v2.2.0 - 16 November 2007, S.J.Kay
// - Make the colour_cont port variable public as is needed for old colour
//   circuit emulation.
// - Port 8 read now returns 0.
//
// v2.1.0 - 20 October 2007, S.J.Kay
// - Implement the modelx information structure.
//
// v1.4.0 - 30 September 2007, S.J.Kay
// - Added vdu_configure() function to set PCG and VDU values for Premium
//   models.
// - Added Premium PC85 ROM model basic upper/lower half select to vdu_lvdat_w
//   function.
// - Removed the non aplpha+ source build option.
// - Change all error output to stdout.
// - Character ROM loading now uses the memmap_loadrom() function.
// - Added changes to error reporting.
//
// v1.0.0 - 21 June 2007, S.J.Kay
// - Added standard and standard+colour machine options to vdu_colcont_w and
//   vdu_lvdat_w functions.
// - Added vdu_lvdat_r and vdu_lvdat_w port handlers.
// - Implemented Alpha+ (Premium) graphics.
//
// v0.0.0 - 8 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.
//==============================================================================

#include <stdio.h>
#include <string.h>
#include <SDL.h>

#include "vdu.h"
#include "crtc.h"
#include "z80.h"
#include "ubee512.h"
#include "memmap.h"
#include "roms.h"
#include "support.h"

#include "macros.h"

//==============================================================================
// structures and variables
//==============================================================================
vdu_t vdu;

int basofs;
static int scrofs;
static int atrofs;
static int pcgofs;
static int colofs;

static int x_colour_cont;
static int lv_dat;              // LV DAT port (0x1c) value
static int x_lv_dat;
static int attribram;           // LV DAT attribute RAM selected if bit 4=1
static int colourram;           // colour RAM selected if bit 6=1 else PCG RAM

extern emu_t emu;
extern model_t modelx;
extern model_custom_t modelc;
extern modio_t modio;
extern crtc_t crtc;

//==============================================================================
// VDU initialise
//
// Make VDU RAM location look like how real static RAM would.  This is needed
// by Basic 5.22e (possibly 6.22e) to detect if hardware supports colour.
// It appears as if 5.22e Basic does not set 0xFFFF to 0xFF in the PCG RAM and
// just assumes it will be non zero from a cold start because of the typical
// static RAM pattern value when starting up the Microbee ?  Running TBASICC.COM
// works when loaded from CP/M because the system already set 0xFFFF to 0xFF
// when it made the inverse character set.
//
// Any unused alpha+ PCG banks will contain random data, testing on a Premium
// Microbee shows the values as mostly zero but can vary greatly between tests.
//
//   pass: void
// return: int                          0 if success else -1
//==============================================================================
int vdu_init (void)
{
 int i;

 // Alpha+ (Premium) variables, 8K for Screen, Colour and attribute, 32K for PCG
 if (modelx.alphap)
    {
     memmap_init6264(vdu.scr_ram, 4);
#if 0
     memmap_init6264(vdu.col_ram, 4);
#else  // prevent b&w checker board on start-up
     memset(vdu.col_ram, 0, sizeof(vdu.col_ram));
#endif
     memmap_init6264(vdu.att_ram, 4);

     crtc.flashbits = 0;
     for (i = 0; i < 16; i++)
        {
         if (i < modelx.pcg)
            memmap_init6264(vdu.pcg_ram + i * 0x0800, 1);
         else
            memset(vdu.pcg_ram + i * 0x0800, 0x00, 0x0800);
        }
    }
 else
    // Standard model 2K Screen, 2K Colour, 2K PCG
    {
     memmap_init6116(vdu.scr_ram, 1);
#if 0
     memmap_init6116(vdu.col_ram, 1);
#else  // prevent b&w checker board on start-up
     memset(vdu.col_ram, 0, sizeof(vdu.col_ram));
#endif
     memmap_init6116(vdu.pcg_ram, 1);
     // Initialise values in the Screen RAM for the PC85 shell ROM
     // and the PCG RAM for Microworld Basic 5.22e
     //
     vdu.pcg_ram[0x07ff] = 0xff;
     vdu.scr_ram[0x043e] = 0xff;
     vdu.scr_ram[0x043f] = 0xff;
    }

 if (modelx.alphap)
    {
     lv_dat = 0;                        // port (0x1c) value
     attribram = 0;                     // attribute RAM select
     crtc.extendram = 0;                // extended graphics select
     basofs = ((lv_dat & B8(00100000)) != 0) * 0x2000;
    }

 if (modelx.colour)
    {
     crtc.colour_cont = 0;
     x_colour_cont = 0;
     colourram = 0;                     // set to PCG
    }

 return 0;
}

//==============================================================================
// VDU de-initialise
//
//   pass: void
// return: int                          0
//==============================================================================
int vdu_deinit (void)
{
 return 0;
}

//==============================================================================
// VDU reset
//
//   pass: void
// return: int                          0
//==============================================================================
int vdu_reset (void)
{
 crtc.latchrom = 0;

 if (modelx.alphap)
    {
     lv_dat = 0;                        // port (0x1c) value
     x_lv_dat = 0;
     attribram = 0;                     // attribute RAM select
     crtc.extendram = 0;                // extended graphics select
     basofs = ((lv_dat & B8(00100000)) != 0) * 0x2000;
    }

 if (modelx.colour)
    {
     crtc.colour_cont = 0;
     x_colour_cont = 0;
     colourram = 0;                     // set to PCG
    }

 return 0;
}

//==============================================================================
// Video memory read.
//
//   pass: uint32_t addr
//         struct z80_memory_read_byte *mem_s
// return: uint8_t
//==============================================================================
uint8_t vdu_vidmem_r (uint32_t addr, struct z80_memory_read_byte *mem_s)
{
#if 0
 if (modio.vdumem)
    log_port_1("vdu_vidmem_r", "addr", port, (unsigned int)addr);
#endif

 if (addr & 0x0800)  // if PCG or Colour RAM
    {
     if (colourram)
        return vdu.col_ram[colofs + (addr & 0x7FF)];
     else
        return vdu.pcg_ram[pcgofs + (addr & 0x7FF)];
    }
 else
    {
     if (! crtc.latchrom)
        {
         if (attribram)
            return vdu.att_ram[atrofs + (addr & 0x7FF)];
         else
            return vdu.scr_ram[scrofs + (addr & 0x7FF)];
        }

     else
/* FIXME this will not work if crtc.disp_start is set to just before 0x2000, and the
   memory addresses would overlap 0x2000 */
        return vdu.chr_rom[((crtc.disp_start & 0x2000) >> 2) + (addr & 0x07FF)];
    }
}

//==============================================================================
// Video memory write.
//
// All writes to VDU memory are first checked to see if the location will
// change. If the location is the same nothing is done, this prevents
// unnecessary time consuming video rendering taking place.
//
//   pass: uint32_t addr
//         uint8_t data
//         struct z80_memory_write_byte *mem_s
// return: void
//==============================================================================
void vdu_vidmem_w (uint32_t addr, uint8_t data, struct z80_memory_write_byte *mem_s)
{
 int addr_x;

#if 0
 if (modio.vdumem)
    log_port_2("vdu_vidmem_w", "addr", "data", port, (unsigned int)addr, data);
#endif

 // if colour or PCG RAM
 if (addr & 0x0800)
    {
     if (! colourram) // if PCG RAM is currently in the video map
        {
         addr_x = pcgofs + (addr & 0x07FF);
         if (vdu.pcg_ram[addr_x] == data)
            return;
         vdu.pcg_ram[addr_x] = data;
         crtc_set_redraw();
        }
     else
        {
         addr_x = colofs + (addr & 0x07FF);
         if (vdu.col_ram[addr_x] == data)
            return;
         vdu.col_ram[addr_x] = data;
         crtc_redraw_char(addr, 0);
        }
    }
 else
    {
     // screen or attribute RAM
     if (! crtc.latchrom)
        {
         addr_x = scrofs + (addr & 0x07FF);
         if (attribram)
            {
             if (vdu.att_ram[addr_x] == data)
                return;
             if ((addr & 0x07FF) < (25 * 80)) // ignore any stuff off screen (possibly BIOS uses to test)
                {
                 if ((vdu.att_ram[addr_x] & B8(10000000)) && (! (data & B8(10000000))) && (crtc.flashbits))
                    crtc.flashbits--;
                 else
                    if ((! (vdu.att_ram[addr_x] & B8(10000000))) && (data & B8(10000000)))
                       crtc.flashbits++;
                }
             vdu.att_ram[addr_x] = data;
             crtc_redraw_char(addr, 0);
            }
         else
            {
             if (vdu.scr_ram[addr_x] == data)
                return;
             vdu.scr_ram[addr_x] = data;
             crtc_redraw_char(addr, 0);
            }
        }
    }
}

//==============================================================================
// Read Port 0x08 - colour control
//
//   pass: uint16_t port
//         struct z80_port_read *port_s
// return: uint16_t                     0
//==============================================================================
uint16_t vdu_colcont_r (uint16_t port, struct z80_port_read *port_s)
{
 if (modio.vdu)
    log_port_0("vdu_colcont_r", port);

 return 0;
}

//==============================================================================
// Write Port 0x08 - colour control
//
// Issue introduced in v3.0.0/1 when running in standard ROM model
// monochrome mode and Telcom versions 1.1 or 1.2.  The top status line
// (after 'BAS') has what appears to be the "emulated" uninitialised screen
// RAM.  Calling crtc_set_redraw() here for non-colour models fixes the
// problem but I'm not sure why or what changes in v3.0.0 caused this issue.
// It appears to be caused by applications that test for a colour model.
//
//   pass: uint16_t port
//         uint8_t data
//         struct z80_port_write *port_s
// return: void
//==============================================================================
void vdu_colcont_w (uint16_t port, uint8_t data, struct z80_port_write *port_s)
{
 if (modelx.colour)
    {
     if (modio.vdu)
        log_port_1("vdu_colcont_w", "data", port, crtc.colour_cont);

     crtc.colour_cont = data & B8(11111110);

     if (modelx.colour == 1)
        {
         // check if any of the RGB background intensity bits changed
         if ((x_colour_cont & B8(00001110)) != (crtc.colour_cont & B8(00001110)))
            crtc_set_redraw();  // all because can affect whole display
        }

     if ((x_colour_cont & B8(01000000)) != (crtc.colour_cont & B8(01000000)))
       colourram = crtc.colour_cont & B8(01000000);

     x_colour_cont = crtc.colour_cont;
    }
 else
    crtc_set_redraw();  // see above for the reason
}

//==============================================================================
// Read Port 0x09 - Colour wait off
//
//   pass: uint16_t port
//         struct z80_port_read *port_s
// return: UNINT16                      0
//==============================================================================
uint16_t vdu_colwait_r (uint16_t port, struct z80_port_read *port_s)
{
 if (modio.vdu)
    log_port_0("vdu_colwait_r", port);

 return 0;
}

//==============================================================================
// Write Port 0x09 - Colour wait off
//
//   pass: uint16_t port
//         uint8_t data
//         struct z80_port_write *port_s
// return: void
//==============================================================================
void vdu_colwait_w (uint16_t port, uint8_t data, struct z80_port_write *port_s)
{
 if (modio.vdu)
    log_port_1("vdu_colwait_w", "data", port, data);
}

//==============================================================================
// Read Port 0x1C - LV DATA
//
// Will only be called if emulating an alpha+ model.
//
//   pass: uint16_t port
//         struct z80_port_read *port_s
// return: uint16_t                     lv_dat
//==============================================================================
uint16_t vdu_lvdat_r (uint16_t port, struct z80_port_read *port_s)
{
 if (modio.vdu)
    log_port_1("vdu_lvdat_r", "lv_dat", port, lv_dat);

 if ((port & 0xff) == 0x1C)
    return lv_dat;
 else
    return 0;
}

//==============================================================================
// Write Port 0x1C - LV DATA
//
// Will only be called if emulating an alpha+ model.
//
//   pass: uint16_t port
//         uint8_t data
//         struct z80_port_write *port_s
// return: void
//==============================================================================
void vdu_lvdat_w (uint16_t port, uint8_t data, struct z80_port_write *port_s)
{
 int videobank;  // LV DAT bank bits 0-3

 if (modio.vdu)
    log_port_1("vdu_lvdat_w", "data", port, data);

 if ((port & 0xff) != 0x1C)
    return;

 lv_dat = data & B8(10111111);                  // port (0x1c) value

 if (x_lv_dat != lv_dat)
    {
     basofs = ((lv_dat & B8(00100000)) != 0) * 0x2000;
     videobank = lv_dat & B8(00001111);         // bank bits 0-3
     attribram = lv_dat & B8(00010000);         // attribute RAM select
     crtc.extendram = lv_dat & B8(10000000);    // extended graphics select

     if (crtc.extendram)        // if the alpha+ graphics are enabled
        {
         scrofs = (videobank & modelx.vdu) * 0x0800;
         atrofs = (videobank & modelx.vdu) * 0x0800;
         colofs = (videobank & modelx.vdu) * 0x0800;
         if (videobank < modelx.pcg)
            pcgofs = videobank * 0x0800;
         else
            pcgofs = 16 * 0x0800;  // use dummy PCB bank (hides)
        }
     else
        {
         scrofs = 0;
         atrofs = 0;
         colofs = 0;
         pcgofs = 0;
        }

     x_lv_dat = lv_dat;                         // port (0x1c) value
     crtc_set_redraw();
    }
}

//==============================================================================
// Port 0x0B - Character ROM select
//
//   pass: uint16_t port
//         uint8_t data
//         struct z80_port_write *port_s
// return: void
//==============================================================================
void vdu_latchrom (uint16_t port, uint8_t data, struct z80_port_write *port_s)
{
 crtc.latchrom = data & B8(00000001);

 if (modio.vdu)
    log_port_1("vdu_latchrom", "crtc.latchrom", port, crtc.latchrom);
}
