//******************************************************************************
//*                                  uBee512                                   *
//*       An emulator for the Microbee Z80 ROM, FDD and HDD based models       *
//*                                                                            *
//*                               video module                                 *
//*                                                                            *
//*                   Copyright (C) 2007-2012 Stewart Kay                      *
//******************************************************************************
//
// Provides SDL and OpenGL video rendering.
//
//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 ROM, FDD and HDD based models.
 *  Copyright (C) 2007-2012 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v5.0.0 - 3 August 2010, K Duckmanton
// - Optimised video_putpixel().  Pixel drawing is now done by a function
//   optimised for the depth of the video surface, resulting in a 2x speed
//   increase.
//
// v4.7.0 - 4 June 2010, S.J.Kay
// - Conditional compilation for MINGW and DARWIN has been removed from the
//   video sections. Testing and branching is now based on an emu.system and
//   emu.sdl_version variable as SDL-1.2.14 fixes some issues in Windows.
//
// v4.5.0 - 1 April 2010, S.J.Kay
// - Changed all occurences of video_t members gl_screen_w and gl_screen_h
//   to gl_imagedim_w and gl_imagedim_h as these were the same values.  The
//   gl_screen_w and gl_screen_h members have been removed from the
//   structure.
//
// v4.2.0 - 13 July 2009, S.J.Kay
// - Prevent mouse cursor enabling if emulating the Microbee mouse.
//
// v4.0.0 - 7 June 2009 S.J.Kay
// - Mac OSX also destroys OpenGL texture when the window is resized.  Make
//   conditionals compile in the same methods used for Windows.
// - Fixed mouse cursor positioning when switching to and from full screen
//   and window displays in video_toggledisplay() function.
//
// v3.1.0 - 22 April 2009, S.J.Kay
// - Removed all occurrences of console_output() function calls.
// - Removed all occurrences of SDL_ShowCursor(SDL_DISABLE), as auto cursor
//   hiding for full screen mode is implemented in gui.c module.
// - Changed video_putpixel() function to ignore negative X and Y values
//   and changed ordering of case statements.
// - Added OSD related code to overlay CRTC and fullscreen cursor changes.
// - Moved commands over from function.c to new video_command() function.
// - Changed all printf() calls to use a local xprintf() function.
// - Added video_gl_filter_update() to allows options to update OpenGL
//   filters immidiately.
// - Added OpenGL conditional code compilation using USE_OPENGL.
//
// v3.0.1 - 13 October 2008, S.J.Kay
// - Toggling full screen under Unices and using SDL video rendering fails
//   due to another variable now used to flag the current full screen state.
//   The video_toggledisplay() function now sets the video.flags.
//
// v3.0.0 - 20 Sepetember 2008, S.J.Kay
// - Created a new file using the video related code from the gui.c module
//   and implement OpenGL texture rendering mode.
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <SDL.h>

#ifdef USE_OPENGL
#include <gl.h>
#endif

#ifdef MINGW
#ifdef USE_OPENGL
#include <glext.h>
#endif
#include <windows.h>
#else
#endif

#include "ubee512.h"
#include "gui.h"
#include "video.h"
#include "crtc.h"
#include "mouse.h"
#include "osd.h"

//==============================================================================
// structures and variables
//==============================================================================
video_t video =
{
 .depth = VIDEO_16,     // default depth used by SDL only
 .type = VIDEO_SDLHW,   // default video rendering mode
#ifdef USE_OPENGL
 .filter_fs = VIDEO_SHARP,
 .filter_win = VIDEO_SOFT,
 .filter_max = VIDEO_SHARP,
 .aspect_bee = VIDEO_ASPECT_BEE,
 .vsync = VIDEO_VSYNC_ON,
 .initial_x_percent = 50
#endif
};

#ifdef USE_OPENGL
static video_gl_t video_gl;
static SDL_Surface *gl_screen;
static int rgb_size[4];
static int ignore_one_resize_event;
#endif

SDL_Surface *screen;
static SDL_VideoInfo video_info;
static SDL_Color colors[256];
video_putpixel_fast_fn video_putpixel_fast_p;

extern emu_t emu;
extern crtc_t crtc;
extern gui_t gui;
extern gui_status_t gui_status;
extern modio_t modio;
extern mouse_t mouse;

#ifdef USE_OPENGL
static int video_gl_values (int crt_w, int crt_h, int win_w, int win_h);
#endif

void video_putpixel_fast_8bpp(int x, int y, int val);
void video_putpixel_fast_16bpp(int x, int y, int val);
void video_putpixel_fast_32bpp(int x, int y, int val);

//==============================================================================
// Video initialise.
//
// Get video information before any SDL_SetVideoMode() is made to save the
// desktop width and height values.
//
// Create an initial window no larger than the maximum X resolution of the
// desktop.
//
//   pass: void
// return: int                  0 if no error, -1 if error
//==============================================================================
int video_init (void)
{
 int win_w;
 int win_h;
 int crt_w;
 int crt_h;

 memcpy(&video_info, SDL_GetVideoInfo(), sizeof(video_info));

 video.desktop_w = video_info.current_w;
 video.desktop_h = video_info.current_h;

#ifdef USE_OPENGL
 if (video.type == VIDEO_GL)
    {
     crtc.yscale = 1;
     crtc.aspect = 1;

     crt_w = VIDEO_WIDTH;
     crt_h = VIDEO_HEIGHT;

     win_w = VIDEO_WIDTH;
     win_h = 0;  // the height will be calculated from the width

     if (video.initial_x_pixels)
        win_w = video.initial_x_pixels;

     if (video.initial_x_percent)
        win_w = (int)(((double)video.desktop_w / 100.0) * video.initial_x_percent);

     if (video.max)
        {
         video.maximised = 1;
         win_w = video.desktop_w;
         win_h = video.desktop_h;
        }
    }
 else
#endif
    {
     crt_w = VIDEO_WIDTH;
     crt_h = VIDEO_HEIGHT;
     win_w = VIDEO_WIDTH;
     win_h = VIDEO_HEIGHT;
     crtc.yscale = crtc.aspect;
    }

 if ((win_w > video.desktop_w) || (video_create_surface(crt_w, crt_h, win_w, win_h) == -1))
    {
     xprintf("video_init: Unable to create X resolution of %d\n", win_w);
     return -1;
    }

 return 0;
}

//==============================================================================
// Video de-initialise.
//
//   pass: void
// return: int                  0
//==============================================================================
int video_deinit (void)
{
 return 0;
}

//==============================================================================
// Video reset.
//
//   pass: void
// return: int                  0
//==============================================================================
int video_reset (void)
{
 return 0;
}

//==============================================================================
// Convert mouse X, Y values to CRTC scaled values.
//
// When using OpenGL mode the mouse X, Y values returned are true screen
// co-ordinates and not the resized values.  This requires the values to be
// converted back to CRTC like values before they can be used.
//
//   pass: int mouse_x                  mouse X position
//         int mouse_y                  mouse Y position
//         int *x                       scaled to CRTC X
//         int *y                       scaled to CRTC Y
// return: void
//==============================================================================
void video_convert_mouse_to_crtc_xy (int mouse_x, int mouse_y, int *x, int *y)
{
#ifdef USE_OPENGL
 if (video.type == VIDEO_GL)
    {
     double crt_w = crtc.hdisp * 8;
     double crt_h = crtc.vdisp * crtc.scans_per_row;
     *x = (int)((crt_w / video.gl_imagedim_w) * mouse_x);
     *y = (int)((crt_h / video.gl_imagedim_h) * mouse_y);
    }
 else
#endif
    {
     *x = mouse_x;
     *y = mouse_y / crtc.yscale;
    }
}

//==============================================================================
// Convert CRTC X, Y values to mouse values.
//
//   pass: int crtc_x                   CRTC X position
//         int crtc_y                   CRTC Y position
//         int *x                       scaled to mouse X
//         int *y                       scaled to mouse Y
// return: void
//==============================================================================
void video_convert_crtc_to_mouse_xy (int crtc_x, int crtc_y, int *x, int *y)
{
#ifdef USE_OPENGL
 if (video.type == VIDEO_GL)
    {
     double crt_w = crtc.hdisp * 8;
     double crt_h = crtc.vdisp * crtc.scans_per_row;
     *x = (int)((crtc_x / crt_w) * video.gl_imagedim_w);
     *y = (int)((crtc_y / crt_h) * video.gl_imagedim_h);
    }
 else
#endif
    {
     *x = crtc_x;
     *y = crtc_y * crtc.yscale;
    }
}

//==============================================================================
// Report video information.
//
//   pass: void
// return: void
//==============================================================================
static void video_report_information (void)
{
 static int reported = 0;

 if ((! modio.video) || (reported++))
    return;

 xprintf("\n");
 xprintf("SDL GENERAL INFORMATION\n");
 xprintf("-----------------------\n");
 xprintf("Desktop width  : %d\n", video.desktop_w);
 xprintf("Desktop height : %d\n", video.desktop_h);
 xprintf("Initial BPP    : %d\n", video_info.vfmt->BitsPerPixel);
 xprintf("Screen BPP     : %d\n", SDL_GetVideoSurface()->format->BitsPerPixel);

#ifdef USE_OPENGL
 int value;

 if (video.type >= VIDEO_GL)
    {
     xprintf("\n");
     xprintf("OPENGL INFORMATION\n");
     xprintf("------------------\n");
     xprintf("Vendor              : %s\n", glGetString(GL_VENDOR));
     xprintf("Renderer            : %s\n", glGetString(GL_RENDERER));
     xprintf("Version             : %s\n", glGetString(GL_VERSION));
#if 0
     xprintf("Extensions          : %s\n", glGetString(GL_EXTENSIONS));
#endif
     xprintf("\n");
     SDL_GL_GetAttribute( SDL_GL_RED_SIZE, &value);
     xprintf("SDL_GL_RED_SIZE     : requested %d, got %d\n", rgb_size[0], value);
     SDL_GL_GetAttribute( SDL_GL_GREEN_SIZE, &value );
     xprintf("SDL_GL_GREEN_SIZE   : requested %d, got %d\n", rgb_size[1], value);
     SDL_GL_GetAttribute( SDL_GL_BLUE_SIZE, &value );
     xprintf("SDL_GL_BLUE_SIZE    : requested %d, got %d\n", rgb_size[2], value);
     SDL_GL_GetAttribute( SDL_GL_ALPHA_SIZE, &value );
     xprintf("SDL_GL_ALPHA_SIZE   : requested %d, got %d\n", rgb_size[3], value);
     SDL_GL_GetAttribute( SDL_GL_DEPTH_SIZE, &value );
     xprintf("SDL_GL_DEPTH_SIZE   : requested %d, got %d\n", video.bpp, value);
     SDL_GL_GetAttribute( SDL_GL_DOUBLEBUFFER, &value );
     xprintf("SDL_GL_DOUBLEBUFFER : requested 1, got %d\n", value);
    }
#endif
 xprintf("\n");
}

#ifdef USE_OPENGL
//==============================================================================
// Calculate values for OpenGL texture rendering method.
//
// This is called from the video_create_surface() function.
//
//   pass: int crt_w            CRT display width
//         int crt_h            CRT display height
//         int win_w            wanted window width
//         int win_h            wanted window height
// return: int                  0 if values are ok, else -1
//==============================================================================
static int video_gl_values (int crt_w, int crt_h, int win_w, int win_h)
{
 double scale_factor;
 double disp_ratio;
 double aspect_adj;

 double ref_w;
 double ref_h;

 int new_w;
 int new_h;

// check the window width wanted
 if (win_w > video.desktop_w)
    return -1;

// calculate the CRT values
 if (! win_w)
    win_w = crtc.hdisp * 8;

 if (! win_h)
    win_h = crtc.vdisp * crtc.scans_per_row;

// Monitor desktop could be 4:3, 5:4 LCD (1280x1024) or widescreen
// If no monitor ratio specified then rely on the resolution to tell us.
// This method should normally work with LCD types and 4:3 CRT monitors
// using 4:3 resolutions.

 if (video.aspect_mon < 0.1)
    disp_ratio = (double)video.desktop_w / video.desktop_h;
 else
    disp_ratio = video.aspect_mon;

 aspect_adj = disp_ratio / video.aspect_bee;

 if (video.fullscreen)
    {
     // calculate resized display values for full screen mode
     ref_w = (double)video.desktop_w;
     ref_h = (double)video.desktop_h;
     new_w = (int)(crt_w * (ref_w / crt_w));
     new_h = (int)(crt_h * (ref_h / crt_h));
    }
 else
    {
     video.last_win_w = win_w;
     video.last_win_h = win_h;

     // if a maximised window then we can not resize the window!, use the values passed
     // and allow any scaling aspect to fill.
     if (video.maximised)
        {
         new_w = win_w;
         new_h = win_h;
        }
     else
        {
         // calculate resized display values
         scale_factor = (double)win_w / video.desktop_w;
         ref_w = video.desktop_w * scale_factor;
         ref_h = video.desktop_h * scale_factor;
         new_w = (int)(crt_w * (ref_w / crt_w));
         new_h = (int)(crt_h * (ref_h / crt_h) * aspect_adj);
        }
    }

 video.gl_basedim_w = crt_w;
 video.gl_basedim_h = crt_h;

 video.gl_imagedim_w = new_w;
 video.gl_imagedim_h = new_h;

// current screen size as a percentage of the desktop width
 video.percent_size = (int)(((double)video.gl_imagedim_w / video.desktop_w) * 100.0);

 return 0;
}

//==============================================================================
// Get one of 3 possible filters depending on the current display type/size
// and return the OpenGL value.
//
//   pass: void
// return: int                  OpenGL filter value
//==============================================================================
static int video_gl_filter_get (void)
{
 int filter;

 if (video.fullscreen)
    filter = video.filter_fs;
 else
    if (video.maximised)
       filter = video.filter_max;
    else
       filter = video.filter_win;

 switch (filter)
    {
     case 0 :
        return GL_LINEAR;
     case 1 :
        return GL_NEAREST;
    }
 return 0;
}

//==============================================================================
// Set OpenGL filter value into the texture.
//
//   pass: int filter           OpenGL filter value
// return: void
//==============================================================================
static void video_gl_filter_set (int filter)
{
 video_gl.filter = filter;
 glBindTexture(GL_TEXTURE_2D, video_gl.texture);
 glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
 glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
 glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, video_gl.filter);
 glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, video_gl.filter);
}

//==============================================================================
// Convert value to a power of two.
//
//   pass: n                    number to be coverted
// return: int                  number converted to a power of two
//==============================================================================
static int video_power_of_two (int n)
{
 int pot = 1;
 while (pot < n)
    pot <<= 1;
 return pot;
}

//==============================================================================
// Create an OpenGL texture.
//
//   pass: void
// return: void
//==============================================================================
static void video_gl_create_texture (void)
{
 if (screen)
     SDL_FreeSurface(screen);

 if (video_gl.texture)
    glDeleteTextures(1, &video_gl.texture);

 memset(&video_gl, 0, sizeof(video_gl));

 video_gl.tex_coord[0] = 0.0f;
 video_gl.tex_coord[1] = 0.0f;

 video_gl.texture_w = video_power_of_two(video.gl_basedim_w);
 video_gl.texture_h = video_power_of_two(video.gl_basedim_h);

 video_gl.tex_coord[2] = (GLfloat)video.gl_basedim_w / video_gl.texture_w;
 video_gl.tex_coord[3] = (GLfloat)video.gl_basedim_h / video_gl.texture_h;

 screen = SDL_CreateRGBSurface(SDL_SWSURFACE,
                               video_gl.texture_w, video_gl.texture_h, 16,
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
                               0x00000000, 0x0000001f, 0x000007e0, 0x0000f800
#else
                               0x0000f800, 0x000007e0, 0x0000001f, 0x00000000
#endif
    );
 if (screen == NULL)
    return;

 video_gl.pixels = screen->pixels;

 switch (screen->format->BytesPerPixel)
    {
     case  2: // 16-bit
        video_gl.pitch = screen->pitch / 2;
        break;
     case  3: // 24-bit
        video_gl.pitch = screen->pitch;
        break;
     case  4: // 32-bit
        video_gl.pitch = screen->pitch / 4;
        break;
     default:
        break;
   }
 switch (screen->format->BytesPerPixel)
    {
     case 4 :
        video_putpixel_fast_p = video_putpixel_fast_32bpp;
        break;
     case 2 :
        video_putpixel_fast_p = video_putpixel_fast_16bpp;
        break;
     case 1 :
        video_putpixel_fast_p = video_putpixel_fast_8bpp;
        break;
    }

// get the filter value for the current display mode
 video_gl.filter = video_gl_filter_get();

 glGenTextures(1, &video_gl.texture);
 glBindTexture(GL_TEXTURE_2D, video_gl.texture);
 glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
 glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
 glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, video_gl.filter);
 glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, video_gl.filter);

#if SDL_BYTEORDER == SDL_BIG_ENDIAN
  video_gl.format = GL_BGR;
  video_gl.type = GL_UNSIGNED_SHORT_5_6_5;
  glTexImage2D(GL_TEXTURE_2D, 0, GL_BGR, video_gl.texture_w, video_gl.texture_h, 0,
  video_gl.format, video_gl.type, video_gl.pixels);
#else
  video_gl.format = GL_RGB;
  video_gl.type = GL_UNSIGNED_SHORT_5_6_5;
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB,
  video_gl.texture_w, video_gl.texture_h, 0, video_gl.format, video_gl.type,
  video_gl.pixels);
#endif
}

//==============================================================================
// Window OpenGL resizing.
//
// Resizes a Window and switches between windowed and fullscreen mode.
//
//   pass: int crt_w            CRT display width
//         int crt_h            CRT display height
//         int win_w            window width resolution wanted
//         int win_h            window height resolution wanted
// return: void
//==============================================================================
static void video_gl_window_resize (int crt_w, int crt_h, int win_w, int win_h)
{
 int filter;

 if (video.fullscreen)
    video.flags |= SDL_FULLSCREEN;
 else
    {
     video.flags &= ~SDL_FULLSCREEN;
     if (! mouse.host_in_use)
        SDL_ShowCursor(SDL_ENABLE);  // show the mouse cursor
    }

 // don't do anything if the values were not accepted
 if (video_gl_values(crt_w, crt_h, win_w, win_h) == -1)
    return;

 if ((gl_screen = SDL_SetVideoMode(video.gl_imagedim_w, video.gl_imagedim_h, 0, video.flags)) == NULL)
    xprintf("video_create_surface: SDL_SetVideoMode failed - %s\n", SDL_GetError());

 glViewport(0, 0, video.gl_imagedim_w, video.gl_imagedim_h);

 glMatrixMode(GL_PROJECTION);
 glLoadIdentity();

 // make the OpenGL display top left 0, 0
 glOrtho(0.0, (double)video.gl_imagedim_w, (double)video.gl_imagedim_h, 0.0, 0.0, 1.0);

 glMatrixMode(GL_MODELVIEW);
 glLoadIdentity();

 // set the filter for the current display mode
 filter = video_gl_filter_get();
 video_gl_filter_set(filter);
}
#endif

//==============================================================================
// Create an SDL or OpenGL surface.
//
// video type   rendering method
// ----------   ----------------
//     0        SDL Software rendering
//     1        SDL Hardware rendering
//     2        OpenGL texture rendering
//
// video depth  colours
// -----------  -------
//     0        SDL 8 bit colour
//     1        SDL 8 bit grey scale
//     2        SDL 16 bit colour
//     3        SDL 32 bit colour
//
//   pass: int crt_w            CRT display width
//         int crt_h            CRT display height
//         int win_w            window width resolution
//         int win_h            window height resolution
// return: int                  0 if no error, -1 if error
//==============================================================================
int video_create_surface (int crt_w, int crt_h, int win_w, int win_h)
{
 int i;

 video.flags = 0;

 if (video.fullscreen)
    video.flags |= SDL_FULLSCREEN;
 else
    {
     if (! mouse.host_in_use)
        SDL_ShowCursor(SDL_ENABLE);  // show the mouse cursor
    }

 switch (video.type)
    {
     case VIDEO_SDLSW : // SDL software rendering
         video.flags |= SDL_SWSURFACE | SDL_ASYNCBLIT;
        break;
     case VIDEO_SDLHW : // SDL hardware rendering
         video.flags |= SDL_HWSURFACE | SDL_DOUBLEBUF | SDL_ASYNCBLIT;
        break;
#ifdef USE_OPENGL
     case VIDEO_GL : // OpenGL texture method
        video.flags |= SDL_OPENGL | SDL_RESIZABLE;
        break;
#endif
    }

 if (video.type < VIDEO_GL)
    {
     crt_h = crt_h * crtc.yscale;

     // SDL video rendering modes
     switch (video.depth)
        {
         case VIDEO_8 :
            video.bpp = 8;
            screen = SDL_SetVideoMode(crt_w, crt_h, 8, video.flags);
            break;
         case VIDEO_8GS :
            video.bpp = 8;
            screen = SDL_SetVideoMode(crt_w, crt_h, 8, video.flags);
            if (screen)
               {
                for (i = 0; i < 256; i++) // create a grey scale
                   {
                    colors[i].r = i;
                    colors[i].g = i;
                    colors[i].b = i;
                   }
                SDL_SetColors(screen, colors, 0, 256);
                SDL_SetPalette(screen, SDL_LOGPAL|SDL_PHYSPAL, colors, 0, 256);
               }
            break;
         case VIDEO_16 :
            video.bpp = 16;
            screen = SDL_SetVideoMode(crt_w, crt_h, 16, video.flags);
            break;
         case VIDEO_32 :
            video.bpp = 32;
            screen = SDL_SetVideoMode(crt_w, crt_h, 32, video.flags);
            break;
        }
     if (screen == NULL)
        {
         xprintf("video_create_surface: SDL_SetVideoMode failed - %s\n", SDL_GetError());
         return -1;
        }
     switch (screen->format->BytesPerPixel)
        {
         case 4 :
            video_putpixel_fast_p = video_putpixel_fast_32bpp;
            break;
         case 2 :
            video_putpixel_fast_p = video_putpixel_fast_16bpp;
            break;
         case 1 :
            video_putpixel_fast_p = video_putpixel_fast_8bpp;
            break;
        }
    }
#ifdef USE_OPENGL
 else
    {
     // OpenGL video rendering modes
     switch (video.type)
        {
         case VIDEO_GL : // OpenGL texture method
            crtc.aspect = 1;  // OpenGL will do all the Y axis stretching

            if (video_info.vfmt->BitsPerPixel <= 8)
               video.bpp = 8;
            else
               video.bpp = video_info.vfmt->BitsPerPixel;

            switch (video.bpp)
               {
                case  8:
                   rgb_size[0] = 3;
                   rgb_size[1] = 3;
                   rgb_size[2] = 2;
                   rgb_size[3] = 0;
                   break;
                case 15:
                case 16:
                   rgb_size[0] = 5;
                   rgb_size[1] = 6;
                   rgb_size[2] = 5;
                   rgb_size[3] = 0;
                   break;
                default:
                   rgb_size[0] = 8;
                   rgb_size[1] = 8;
                   rgb_size[2] = 8;
                   rgb_size[3] = 0;
                   break;
               }

            // don't do anything if the values were not accepted
            if (video_gl_values(crt_w, crt_h, win_w, win_h) == -1)
               return -1;

            SDL_GL_SetAttribute(SDL_GL_RED_SIZE, rgb_size[0]);
            SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, rgb_size[1]);
            SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, rgb_size[2]);
            SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE, rgb_size[3]);
            SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
            SDL_GL_SetAttribute(SDL_GL_ACCELERATED_VISUAL, 1);

            // Vsync: swap buffers every n'th retrace (0 to disable, 1 is the default)
            SDL_GL_SetAttribute(SDL_GL_SWAP_CONTROL, video.vsync);

            // don't free this surface if resetting!
            if ((gl_screen = SDL_SetVideoMode(video.gl_imagedim_w, video.gl_imagedim_h, 0, video.flags)) == NULL)
               {
                xprintf("video_create_surface: SDL_SetVideoMode failed - %s\n", SDL_GetError());
                return -1;
               }

            // initialise the OpenGL display
            glViewport(0, 0, video.gl_imagedim_w, video.gl_imagedim_h);
            glShadeModel(GL_FLAT);
            glDisable(GL_CULL_FACE);
            glDisable(GL_DEPTH_TEST);
            glDisable(GL_ALPHA_TEST);
            glDisable(GL_LIGHTING);
            glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);

            glMatrixMode(GL_PROJECTION);
            glLoadIdentity();

            // make the OpenGL display top left 0, 0
            glOrtho(0.0, (double)video.gl_imagedim_w, (double)video.gl_imagedim_h, 0.0, 0.0, 1.0);

            glMatrixMode(GL_MODELVIEW);
            glLoadIdentity();

            // create an OpenGL texture
            video_gl_create_texture();

            glEnable(GL_TEXTURE_2D);

            // clear the display and update it
            glClear(GL_COLOR_BUFFER_BIT);
            SDL_GL_SwapBuffers();

            // clear the current display buffer
            glClear(GL_COLOR_BUFFER_BIT);
            break;
        }
    }
#endif
 video_report_information();  // report video information

 return 0;
}

//==============================================================================
// Video renderer.
//
// video type   rendering method
// ----------   ----------------
//     0        SDL Software rendering
//     1        SDL Hardware rendering
//     2        OpenGL texture rendering
//
//   pass: void
// return: void
//==============================================================================
void video_renderer (void)
{
 switch (video.type)
    {
     case VIDEO_SDLSW : // SDL software rendering
        SDL_UpdateRect(screen, 0, 0, 0, 0);
        break;
     case VIDEO_SDLHW : // SDL hardware rendering
        SDL_Flip(screen);
        break;
#ifdef USE_OPENGL
     case VIDEO_GL : // OpenGL texture method
        glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, video_gl.texture_w,
        video_gl.texture_h, video_gl.format, video_gl.type,
        video_gl.pixels);

        glBegin(GL_QUADS);
        glTexCoord2f(video_gl.tex_coord[0], video_gl.tex_coord[1]);
        glVertex2i(0, 0);
        glTexCoord2f(video_gl.tex_coord[2], video_gl.tex_coord[1]);
        glVertex2i(video.gl_imagedim_w, 0);
        glTexCoord2f(video_gl.tex_coord[2], video_gl.tex_coord[3]);
        glVertex2i(video.gl_imagedim_w, video.gl_imagedim_h);
        glTexCoord2f(video_gl.tex_coord[0], video_gl.tex_coord[3]);
        glVertex2i(0, video.gl_imagedim_h);
        glEnd();

        SDL_GL_SwapBuffers();
        break;
#endif
    }
}

#ifdef USE_OPENGL
//==============================================================================
// Resize or create a new surface depending on the platform in use.
//
//   pass: int crt_w            CRT display width
//         int crt_h            CRT display height
//         int win_w            wanted window width
//         int win_h            wanted window height
// return: int                  0 if no error, -1 if error
//==============================================================================
int video_gl_surface (int crt_w, int crt_h, int win_w, int win_h)
{
 if (((emu.system & EMU_SYSTEM_WINDOWS) && (emu.sdl_version >= 1020014)) || (emu.system & EMU_SYSTEM_UNIX))
    {
     video_gl_window_resize(crt_w, crt_h, win_w, win_h);
     if (emu.display_context == EMU_OSD_CONTEXT)
        {
         osd_redraw();
         video_renderer();
        }
     return 0;
    }

 if (video_create_surface(crt_w, crt_h, win_w, win_h) == -1)
    return -1;
 crtc_set_redraw();
 crtc_redraw();
 if (emu.display_context == EMU_OSD_CONTEXT)
    osd_redraw();
 video_renderer();
 return 0;
}

//==============================================================================
// Window resize event handler.
//
// When changing the window size manually and the dimensions exceed or equal
// the maximum window size a resize event may be issued.  Those places that
// do this can set 'ignore_one_resize_event' equal to 1 to ignore the event
// and preventing it interfering with the manual change.
//
// The 'ignore_one_resize_event' variable is cleared on each video_update()
// call to prevent normal resize events being missed.  This may cancel out
// any ignore request but won't matter when scrolling window sizes.
//
//   pass: void
// return: void
//==============================================================================
void video_gl_resize_event (void)
{
 int crt_w;
 int crt_h;

 if (video.type != VIDEO_GL)
    return;

// ignore one resize event caused when manually changing the display
 if (ignore_one_resize_event)
    {
     ignore_one_resize_event = 0;
     return;
    }

 crt_w = crtc.hdisp * 8;
 crt_h = crtc.vdisp * crtc.scans_per_row;

 video.maximised = (emu.event.resize.w == video.desktop_w);

 if (video.maximised)
    {
     video.maximised_w = emu.event.resize.w;
     video.maximised_h = emu.event.resize.h;
    }

 video_gl_surface(crt_w, crt_h, emu.event.resize.w, emu.event.resize.h);

 gui_status_set_persist(GUI_PERSIST_WIN, 0);
 gui_status_update();
}

//==============================================================================
// Window size setting by command.
//
//   pass: int p                p is percentage in multiples of 10% of the
//                              desktop width or 0 to use current CRT width
//                              size.
// return: void
//==============================================================================
void video_gl_size_set (int p)
{
 int crt_w;
 int crt_h;
 int win_w;

 if ((video.type != VIDEO_GL) || (video.maximised))
    return;

 video.fullscreen = 0;

 crt_w = crtc.hdisp * 8;
 crt_h = crtc.vdisp * crtc.scans_per_row;

 if (p == 0)
    {
     // set window size equal to the current CRT size
     video_gl_surface(crt_w, crt_h, crt_w, crt_h);
    }
 else
    if ((p > 0) && (p < 10))
       {
        // set window size by percentage of desktop width
        win_w = (int)(((double)video.desktop_w / 10.0) * p);
        video_gl_surface(crt_w, crt_h, win_w, 0);
       }
}

//==============================================================================
// Window size setting by incrementing a percentage value.
//
//   pass: int dir              'i' for increase, 'd' for decrease
// return: void
//==============================================================================
void video_gl_size_increment (int dir)
{
 int crt_w;
 int crt_h;
 int win_w;

 if ((video.type != VIDEO_GL) || video.fullscreen || video.maximised)
    return;

// make the next resize event be ignored
 ignore_one_resize_event = 1;

 crt_w = crtc.hdisp * 8;
 crt_h = crtc.vdisp * crtc.scans_per_row;

 if (dir == 'i')
    {
     video.percent_size += VIDEO_INCREMENT_PERCENT;
     if (video.percent_size > VIDEO_MAX_PERCENT)
        video.percent_size = VIDEO_MAX_PERCENT;
    }
 else
    {
     video.percent_size -= VIDEO_INCREMENT_PERCENT;
     if (video.percent_size < VIDEO_MIN_PERCENT)
        video.percent_size = VIDEO_MIN_PERCENT;
    }

 win_w = (int)(((double)video.desktop_w / 100.0) * video.percent_size);
 video_gl_surface(crt_w, crt_h, win_w, 0);
}

//==============================================================================
// Toggle OpenGL (textured) filter mode.  Provides soft and sharp display
// rendering for each of the 3 possible filters depending on the current
// display type/size.
//
//   pass: void
// return: void
//==============================================================================
void video_gl_filter_toggle (void)
{
 int filter;

 if (video.type != VIDEO_GL)
    return;

 if (video.fullscreen)
    video.filter_fs = ! video.filter_fs;
 else
    if (video.maximised)
       video.filter_max = ! video.filter_max;
    else
       video.filter_win = ! video.filter_win;

 filter = video_gl_filter_get();
 video_gl_filter_set(filter);

 crtc_redraw();
 if (emu.display_context == EMU_OSD_CONTEXT)
    osd_redraw();

 video_renderer();
}

//==============================================================================
// Set the aspect ratio of the Microbee window.
//
//   pass: float aspect
// return: int                          0 if no error, -1 if error
//==============================================================================
int video_gl_aspect_bee (float aspect)
{
 if ((aspect < 0.1) || (aspect > 10.0))
    return -1;

 video.aspect_bee = aspect;
 return 0;
}

//==============================================================================
// Set the aspect ratio of the host monitor.
//
//   pass: float aspect
// return: int                          0 if no error, -1 if error
//==============================================================================
int video_gl_aspect_mon (float aspect)
{
 if ((aspect < 0.1) || (aspect > 10.0))
    return -1;

 video.aspect_mon = aspect;
 return 0;
}

//==============================================================================
// Set initial window size value from a set number of pixels wide.  If the
// value is large i.e. side window borderings + width => desktop width than
// SDL will place the window into maximised window mode.
//
//   pass: int pixels                   number of piexels wide
// return: int                          0 if no error, -1 if error
//==============================================================================
int video_gl_size_pixels (int pixels)
{
 if (pixels < 50)
    return -1;

 video.initial_x_pixels = pixels;
 video.initial_x_percent = 0;
 return 0;
}

//==============================================================================
// Set initial window size value from a percentage value.  If the value
// is large i.e. side window borderings + width => desktop width than SDL
// will place the window into maximised window mode.
//
//   pass: int pixels                   number of piexels wide
// return: int                          0 if no error, -1 if error
//==============================================================================
int video_gl_size_percent (int percent)
{
 if ((percent < 5) || (percent > 100))
    return -1;

 video.initial_x_percent = percent;
 video.initial_x_pixels = 0;
 return 0;
}

//==============================================================================
// Update the OpenGL filter to the current values.  This is intended to be
// called from the options during run mode.
//
//   pass: int pixels                   number of piexels wide
// return: void
//==============================================================================
void video_gl_filter_update (void)
{
 if ((emu.runmode) && (video.type == VIDEO_GL))
    {
     video_gl_filter_set(video_gl_filter_get());

     crtc_redraw();
     if (emu.display_context == EMU_OSD_CONTEXT)
        osd_redraw();

     video_renderer();
    }
}
#endif

//==============================================================================
// Write a pixel to the display buffer.
//
// If X or Y co-ordinates are out of range then the function returns without
// doing anything.  Out of range co-ordinates can be considered normal when
// OSD dialogues are dragged off screen.
//
//   pass: int x
//         int y
//         uint32_t col
// return: void
//==============================================================================
void video_putpixel (int x, int y, int col)
{
 if ((x < 0) || (x >= screen->w) || (y < 0) || (y >= screen->h))
    return;
 video_putpixel_fast(x, y, col);
}

//==============================================================================
// "Fast" pixel drawing functions.
//==============================================================================
void video_putpixel_fast_8bpp(int x, int y, int val)
{
    void *p = screen->pixels + y * screen->pitch + x * 1;
    *(uint8_t *)p = val;
}
void video_putpixel_fast_16bpp(int x, int y, int val)
{
    void *p = screen->pixels + y * screen->pitch + x * 2;
    *(uint16_t *)p = val;
}
void video_putpixel_fast_32bpp(int x, int y, int val)
{
    void *p = screen->pixels + y * screen->pitch + x * 4;
    *(uint32_t *)p = val;
}


//==============================================================================
// Video toggle display mode.
//
// The SDL_WM_ToggleFullScreen(screen) is only for X11 at present according
// to the documentation, the win32 build and OpenGL modes uses other
// methods to do this.
//
//   pass: void
// return: int                  0 if no error, -1 if error
//==============================================================================
int video_toggledisplay (void)
{
 int mouse_x;
 int mouse_y;

 int x;
 int y;

 int crt_w;
 int crt_h;

 crt_w = crtc.hdisp * 8;
 crt_h = crtc.vdisp * crtc.scans_per_row;

 video.fullscreen = (! video.fullscreen);

 // get the current mouse X, Y values
 SDL_GetMouseState(&mouse_x, &mouse_y);
 video_convert_mouse_to_crtc_xy(mouse_x, mouse_y, &x, &y);

#ifdef USE_OPENGL
 if (video.type == VIDEO_GL)
    {
     // this should work with Windows and SDL 1.2.14 but it doesn't and
     // results in an all white display so for now Windows still has to
     // recreate the display and we will assume the same for Darwin too.
     // (see below).
     if (emu.system & EMU_SYSTEM_UNIX)
        {
         video_gl_window_resize(crt_w, crt_h, video.last_win_w, video.last_win_h);
         crtc_set_redraw();
         crtc_redraw();
         if (emu.display_context == EMU_OSD_CONTEXT)
            osd_redraw();
         video_renderer();
         gui_changed_videostate();
         video_convert_crtc_to_mouse_xy(x, y, &mouse_x, &mouse_y);
         SDL_WarpMouse(mouse_x, mouse_y);
         return 0;
        }

     // comes here if a Windows or Darwin system.
     // won't switch back correctly to maximised, uses minimised when a new
     // surface is created so we prevent it for win32.
     if (video.maximised && video.fullscreen)
        {
         video.fullscreen = 0;
         return 0;
        }

     if (video_create_surface(crt_w, crt_h, video.last_win_w, video.last_win_h) == -1)
        return -1;

     crtc_set_redraw();
     crtc_redraw();
     if (emu.display_context == EMU_OSD_CONTEXT)
        osd_redraw();
     video_renderer();
     gui_changed_videostate();
     video_convert_crtc_to_mouse_xy(x, y, &mouse_x, &mouse_y);
     SDL_WarpMouse(mouse_x, mouse_y);
     return 0;
    }
 else
#endif
    {
     if (! (emu.system & EMU_SYSTEM_WINDOWS))
        {
         if (video.fullscreen)
            video.flags |= SDL_FULLSCREEN;
         else
            {
             video.flags &= ~SDL_FULLSCREEN;
             if (! mouse.host_in_use)
                SDL_ShowCursor(SDL_ENABLE);  // show the mouse cursor
            }
         SDL_WM_ToggleFullScreen(screen);
         gui_changed_videostate();
         video_convert_crtc_to_mouse_xy(x, y, &mouse_x, &mouse_y);
         SDL_WarpMouse(mouse_x, mouse_y);
         return 0;
        }

     if (video_create_surface(crt_w, crt_h, 0, 0) == -1)
        return -1;

     crtc_set_redraw();
     crtc_redraw();
     if (emu.display_context == EMU_OSD_CONTEXT)
        osd_redraw();
     video_renderer();
     gui_changed_videostate();
     video_convert_crtc_to_mouse_xy(x, y, &mouse_x, &mouse_y);
     SDL_WarpMouse(mouse_x, mouse_y);
     return 0;
    }
}

//==============================================================================
// Video update. This is called after each Z80 code frame has completed.
//
// Redraws the surface then updates the display if required. The crtc.update
// flag greatly reduces host CPU time.
//
//   pass: void
// return: void
//==============================================================================
void video_update (void)
{
 osd_update();          // sets the crtc.update flag if OSD needs refreshing

 crtc_redraw();         // only redraws if corresponding flag is set.

#ifdef USE_OPENGL
// re-enable resize events after changing window size manually.
 ignore_one_resize_event = 0;
#endif

 if (crtc.update)
    {
     if (emu.display_context == EMU_OSD_CONTEXT)
        osd_redraw();
     video_renderer();
     crtc.update = 0;
    }
}

//==============================================================================
// Video commands.
//
//   pass: int cmd                      video command
//         int p                        parameter
// return: void
//==============================================================================
void video_command (int cmd, int p)
{
 switch (cmd)
    {
     case EMU_CMD_FULLSCR :
        video_toggledisplay();
        break;
     case EMU_CMD_SCREENI :
#ifdef USE_OPENGL
        if (video.type != VIDEO_GL)
           break;
        if ((gui_status.win) || (gui.persist_flags & GUI_PERSIST_WIN))
           video_gl_size_increment('i');
        if (! gui_status.win)
           gui_status_set_persist(GUI_PERSIST_WIN, 0);
#endif
        break;
     case EMU_CMD_SCREEND :
#ifdef USE_OPENGL
        if (video.type != VIDEO_GL)
           break;
        if ((gui_status.win) || (gui.persist_flags & GUI_PERSIST_WIN))
           video_gl_size_increment('d');
        if (! gui_status.win)
           gui_status_set_persist(GUI_PERSIST_WIN, 0);
#endif
        break;
     case EMU_CMD_VIDSIZE1 :
#ifdef USE_OPENGL
        if (video.type != VIDEO_GL)
           break;
        video_gl_size_set(p);
        if (! gui_status.win)
           gui_status_set_persist(GUI_PERSIST_WIN, 0);
#endif
        break;
     case EMU_CMD_GL_FILTER :
#ifdef USE_OPENGL
        if (video.type != VIDEO_GL)
           break;
        video_gl_filter_toggle();
#endif
        break;
    }
}
